# Copyright (c) 2015, MapR Technologies
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import collections as c

import sahara.plugins.provisioning as p
import sahara.plugins.utils as u
from sahara_plugin_mapr.i18n import _
import sahara_plugin_mapr.plugins.mapr.abstract.version_handler as vh
import sahara_plugin_mapr.plugins.mapr.base.base_cluster_configurer as b_conf
import sahara_plugin_mapr.plugins.mapr.base.base_cluster_validator as bv
import sahara_plugin_mapr.plugins.mapr.base.base_edp_engine as edp
import sahara_plugin_mapr.plugins.mapr.base.base_health_checker as health
import sahara_plugin_mapr.plugins.mapr.base.base_node_manager as bs
from sahara_plugin_mapr.plugins.mapr import images
import sahara_plugin_mapr.plugins.mapr.util.general as util


class BaseVersionHandler(vh.AbstractVersionHandler):
    def __init__(self):
        self._validator = bv.BaseValidator()
        self._configurer = b_conf.BaseConfigurer()
        self._health_checker = health.BaseHealthChecker()
        self._node_manager = bs.BaseNodeManager()
        self._version = None
        self._required_services = []
        self._services = []
        self._node_processes = {}
        self._configs = []
        self.images = images

    def get_edp_engine(self, cluster, job_type):
        if job_type in edp.MapROozieJobEngine.get_supported_job_types():
            return edp.MapROozieJobEngine(cluster)
        return None

    def get_edp_job_types(self):
        return edp.MapROozieJobEngine.get_supported_job_types()

    def get_edp_config_hints(self, job_type):
        return edp.MapROozieJobEngine.get_possible_job_config(job_type)

    def get_services(self):
        return self._services

    def get_required_services(self):
        return self._required_services

    def get_node_processes(self):
        if not self._node_processes:
            self._node_processes = {
                s.ui_name: [np.ui_name for np in s.node_processes]
                for s in self.get_services() if s.node_processes}
        return self._node_processes

    def get_configs(self):
        if not self._configs:
            configs = [c for s in self.get_services() for c in s.get_configs()]
            configs += self._get_version_configs()
            configs += self._get_repo_configs()
            self._configs = util.unique_list(configs)
        return self._configs

    def _get_repo_configs(self):
        ubuntu_base = p.Config(
            name="Ubuntu base repo",
            applicable_target="general",
            scope='cluster',
            priority=1,
            default_value="",
            description=_(
                'Specifies Ubuntu MapR core repository.')
        )
        centos_base = p.Config(
            name="CentOS base repo",
            applicable_target="general",
            scope='cluster',
            priority=1,
            default_value="",
            description=_(
                'Specifies CentOS MapR core repository.')
        )
        ubuntu_eco = p.Config(
            name="Ubuntu ecosystem repo",
            applicable_target="general",
            scope='cluster',
            priority=1,
            default_value="",
            description=_(
                'Specifies Ubuntu MapR ecosystem repository.')
        )
        centos_eco = p.Config(
            name="CentOS ecosystem repo",
            applicable_target="general",
            scope='cluster',
            priority=1,
            default_value="",
            description=_(
                'Specifies CentOS MapR ecosystem repository.')
        )
        return [ubuntu_base, centos_base, ubuntu_eco, centos_eco]

    def _get_version_configs(self):
        services = self.get_services()

        service_version_dict = c.defaultdict(list)
        for service in services:
            service_version_dict[service.ui_name].append(service.version)

        result = []
        for service in services:
            versions = service_version_dict[service.ui_name]
            if len(versions) > 1:
                result.append(service.get_version_config(versions))

        return result

    def get_configs_dict(self):
        configs = dict()
        for service in self.get_services():
            configs.update(service.get_configs_dict())
        return configs

    def configure_cluster(self, cluster):
        instances = u.get_instances(cluster)
        cluster_context = self.get_context(cluster, added=instances)
        self._configurer.configure(cluster_context)

    def start_cluster(self, cluster):
        instances = u.get_instances(cluster)
        cluster_context = self.get_context(cluster, added=instances)
        self._node_manager.start(cluster_context)
        self._configurer.post_start(cluster_context)

    def validate(self, cluster):
        cluster_context = self.get_context(cluster)
        self._validator.validate(cluster_context)

    def validate_scaling(self, cluster, existing, additional):
        cluster_context = self.get_context(cluster)
        self._validator.validate_scaling(cluster_context, existing, additional)

    def scale_cluster(self, cluster, instances):
        cluster_context = self.get_context(cluster, added=instances)
        cluster_context._cluster_services = None
        self._configurer.configure(cluster_context, instances)
        self._configurer.update(cluster_context, instances)
        self._node_manager.start(cluster_context, instances)

    def decommission_nodes(self, cluster, instances):
        cluster_context = self.get_context(cluster, removed=instances)
        cluster_context._cluster_services = None
        self._node_manager.move_nodes(cluster_context, instances)
        self._node_manager.stop(cluster_context, instances)
        self._node_manager.await_no_heartbeat()
        self._node_manager.remove_nodes(cluster_context, instances)
        self._configurer.update(cluster_context, instances)

    def get_open_ports(self, node_group):
        result = []
        for service in self.get_services():
            for node_process in service.node_processes:
                if node_process.ui_name in node_group.node_processes:
                    result += node_process.open_ports
        return util.unique_list(result)

    def get_cluster_checks(self, cluster):
        cluster_context = self.get_context(cluster)
        return self._health_checker.get_checks(cluster_context)

    def get_image_arguments(self):
        if hasattr(self, 'images'):
            return self.images.get_image_arguments()
        else:
            return NotImplemented

    def pack_image(self, hadoop_version, remote, test_only=False,
                   image_arguments=None):
        if hasattr(self, 'images'):
            self.images.pack_image(
                remote, test_only=test_only, image_arguments=image_arguments)

    def validate_images(self, cluster, test_only=False, image_arguments=None):
        if hasattr(self, 'images'):
            self.images.validate_images(
                cluster, test_only=test_only, image_arguments=image_arguments)



# Copyright (c) 2015, MapR Technologies
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import sahara.plugins.utils as utils
import sahara_plugin_mapr.plugins.mapr.domain.configuration_file as bcf
import sahara_plugin_mapr.plugins.mapr.domain.node_process as np
import sahara_plugin_mapr.plugins.mapr.domain.service as s
import sahara_plugin_mapr.plugins.mapr.services.hbase.hbase as hbase
import sahara_plugin_mapr.plugins.mapr.services.hive.hive as hive
import sahara_plugin_mapr.plugins.mapr.services.sentry.sentry as sentry
import sahara_plugin_mapr.plugins.mapr.util.general as g
import sahara_plugin_mapr.plugins.mapr.util.validation_utils as vu

IMPALA_SERVER = np.NodeProcess(
    name='impalaserver',
    ui_name='Impala-Server',
    package='mapr-impala-server',
    open_ports=[21000, 21050, 25000]
)
IMPALA_STATE_STORE = np.NodeProcess(
    name='impalastore',
    ui_name='Impala-Statestore',
    package='mapr-impala-statestore',
    open_ports=[25010]
)
IMPALA_CATALOG = np.NodeProcess(
    name='impalacatalog',
    ui_name='Impala-Catalog',
    package='mapr-impala-catalog',
    open_ports=[25020]
)


class Impala(s.Service):
    def __init__(self):
        super(Impala, self).__init__()
        self._name = 'impala'
        self._ui_name = 'Impala'
        self._node_processes = [
            IMPALA_CATALOG,
            IMPALA_SERVER,
            IMPALA_STATE_STORE,
        ]

    def _get_impala_env_props(self, cluster_context):
        return {}

    def post_start(self, cluster_context, instances):
        self._copy_hive_site(cluster_context, instances)

    def _copy_hive_site(self, cluster_context, instances):
        hive_site_path = self._hive(cluster_context).conf_dir(
            cluster_context) + "/hive-site.xml"
        path = self.conf_dir(cluster_context) + "/hive-site.xml"
        hive_instance = cluster_context.get_instance(hive.HIVE_METASTORE)
        impalas = cluster_context.filter_instances(instances, service=self)
        for instance in impalas:
            g.copy_file(hive_site_path, hive_instance, path, instance,
                        run_as='root', owner='mapr')

    # hive service instance
    def _hive(self, cluster_context):
        hive_version = cluster_context.get_chosen_service_version('Hive')
        return cluster_context._find_service_instance('Hive', hive_version)

    def get_config_files(self, cluster_context, configs, instance=None):
        defaults = 'plugins/mapr/services/impala/resources/impala-env.sh.j2'

        impala_env = bcf.TemplateFile("env.sh")
        impala_env.remote_path = self.conf_dir(cluster_context)
        if instance:
            impala_env.fetch(instance)
        impala_env.parse(utils.get_file_text(defaults, 'sahara_plugin_mapr'))
        impala_env.add_properties(self._get_impala_env_props(cluster_context))
        sentry_host = cluster_context.get_instance(sentry.SENTRY)
        if sentry_host:
            sentry_mode = cluster_context._get_cluster_config_value(
                sentry.Sentry().SENTRY_STORAGE_MODE)
            ui_name = sentry.Sentry().ui_name
            sentry_version = cluster_context.get_chosen_service_version(
                ui_name)
            sentry_service = cluster_context. \
                _find_service_instance(ui_name, sentry_version)
            if sentry_service.supports(self, sentry_mode):
                impala_env.add_properties({
                    'sentry_home': sentry_service.home_dir(cluster_context),
                    'sentry_db': sentry_mode == sentry.DB_STORAGE_SENTRY_MODE,
                    'sentry_policy_file': 'maprfs://'
                                          + sentry_service.GLOBAL_POLICY_FILE,
                })
        return [impala_env]

    def post_install(self, cluster_context, instances):
        self._set_service_dir_owner(cluster_context, instances)


class ImpalaV141(Impala):
    def __init__(self):
        super(ImpalaV141, self).__init__()
        self._version = '1.4.1'
        self._dependencies = [
            ('mapr-hive', hive.HiveV013().version),
            ('mapr-impala', self.version),
        ]
        self._validation_rules = [
            vu.depends_on(hive.HiveV013(), self),
            vu.exactly(1, IMPALA_STATE_STORE),
            vu.exactly(1, IMPALA_CATALOG),
            vu.at_least(1, IMPALA_SERVER),
        ]

    def _get_impala_env_props(self, cluster_context):
        return {
            'impala_version': self.version,
            'statestore_host': cluster_context.get_instance_ip(
                IMPALA_STATE_STORE),
            'catalog_host': cluster_context.get_instance_ip(IMPALA_CATALOG),
        }


class ImpalaV220(Impala):
    def __init__(self):
        super(ImpalaV220, self).__init__()
        self._version = '2.2.0'
        self._dependencies = [
            ('mapr-hive', hive.HiveV12().version),
            ('mapr-impala', self.version),
        ]
        self._validation_rules = [
            vu.depends_on(hive.HiveV12(), self),
            vu.exactly(1, IMPALA_STATE_STORE),
            vu.exactly(1, IMPALA_CATALOG),
            vu.at_least(1, IMPALA_SERVER),
            vu.required_os('centos', self)
        ]

    def _get_impala_env_props(self, cluster_context):
        return {
            'impala_version': self.version,
            'statestore_host': cluster_context.get_instance_ip(
                IMPALA_STATE_STORE),
            'catalog_host': cluster_context.get_instance_ip(IMPALA_CATALOG),
        }

    def _get_packages(self, cluster_context, node_processes):
        result = []

        result += self.dependencies
        result += [(np.package, self.version) for np in node_processes]
        # gets the latest version
        hbase_version = cluster_context.get_chosen_service_version('HBase')
        result += [('mapr-hbase', hbase_version)]

        return result


class ImpalaV250(Impala):
    def __init__(self):
        super(ImpalaV250, self).__init__()
        self._version = '2.5.0'
        self._dependencies = [
            ('mapr-hive', hive.HiveV12().version),
            ('mapr-impala', self.version),
            ('mapr-hbase', hbase.HBaseV111().version)
        ]
        self._validation_rules = [
            vu.depends_on(hive.HiveV12(), self),
            vu.exactly(1, IMPALA_STATE_STORE),
            vu.exactly(1, IMPALA_CATALOG),
            vu.at_least(1, IMPALA_SERVER),
            vu.required_os('centos', self)
        ]

    def _get_impala_env_props(self, cluster_context):
        return {
            'impala_version': self.version,
            'statestore_host': cluster_context.get_instance_ip(
                IMPALA_STATE_STORE),
            'catalog_host': cluster_context.get_instance_ip(IMPALA_CATALOG),
        }



# Copyright (c) 2015, MapR Technologies
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import oslo_serialization.jsonutils as json

from sahara.plugins import utils as plugin_utils
from sahara_plugin_mapr.i18n import _
import sahara_plugin_mapr.plugins.mapr.util.general as util


WARDEN_MANAGED_CMD = ('sudo -u mapr maprcli node services'
                      ' -name %(service)s'
                      ' -action %(action)s'
                      ' -nodes %(nodes)s')


class NodeProcess(object):
    def __init__(self, name, ui_name, package, open_ports=None):
        self._name = name
        self._ui_name = ui_name
        self._package = package
        self._open_ports = open_ports or []

    @property
    def name(self):
        return self._name

    @property
    def ui_name(self):
        return self._ui_name

    @property
    def package(self):
        return self._package

    @property
    def open_ports(self):
        return self._open_ports

    def start(self, instances):
        self.execute_action(instances, Action.START)

    def restart(self, instances):
        self.execute_action(instances, Action.RESTART)

    def stop(self, instances):
        self.execute_action(instances, Action.STOP)

    def execute_action(self, instances, action):
        if len(instances) == 0:
            return
        nodes = ','.join(map(lambda i: i.internal_ip, instances))
        args = {'service': self.name, 'action': action.name, 'nodes': nodes}
        command = WARDEN_MANAGED_CMD % args
        with instances[0].remote() as r:
            r.execute_command(command)
        self._wait_for_status(instances, action.status)

    def _wait_for_status(self, instances, status, sleep=3, timeout=60):
        def poll_status(instance):
            operation_name = _('Wait for {node_process} on {instance}'
                               ' to change status to "{status}"')
            args = {
                'node_process': self.ui_name,
                'instance': instance.instance_name,
                'status': status.name,
            }
            return plugin_utils.poll(
                get_status=lambda: self.status(instance) == status,
                operation_name=operation_name.format(**args),
                timeout=timeout,
                sleep=sleep,
            )

        util.execute_on_instances(instances, poll_status)

    def status(self, instance):
        command = 'maprcli service list -node %s -json' % instance.internal_ip
        with instance.remote() as remote:
            ec, out = remote.execute_command(util._run_as('mapr', command))
        node_processes = json.loads(out)['data']
        for node_process in node_processes:
            if node_process['name'] == self.name:
                return Status.by_value(node_process['state'])

        return Status.NOT_CONFIGURED

    def is_started(self, instance):
        # At least tried to do it =)
        return self.status(instance) in [Status.RUNNING,
                                         Status.FAILED,
                                         Status.STAND_BY]


class Status(object):
    class Item(object):
        def __init__(self, name, value):
            self._name = name
            self._value = value

        @property
        def name(self):
            return self._name

        @property
        def value(self):
            return self._value

    # The package for the service is not installed and/or
    # the service is not configured (configure.sh has not run)
    NOT_CONFIGURED = Item('Not Configured', 0)

    # The package for the service is installed and configured
    CONFIGURED = Item('Configured', 1)

    # The service is installed, started by the warden, and is currently running
    RUNNING = Item('Running', 2)

    # The service is installed and configure.sh has run,
    # but the service is not running
    STOPPED = Item('Stopped', 3)

    # The service is installed and configured, but not running
    FAILED = Item('Failed', 4)

    # The service is installed and is in standby mode, waiting to take over
    # in case of failure of another instance.
    # Mainly used for JobTracker warm standby
    STAND_BY = Item('Standby', 5)

    @staticmethod
    def items():
        return [
            Status.NOT_CONFIGURED,
            Status.CONFIGURED,
            Status.RUNNING,
            Status.STOPPED,
            Status.FAILED,
            Status.STAND_BY,
        ]

    @staticmethod
    def by_value(value):
        for v in Status.items():
            if v.value == value:
                return v


class Action(object):
    class Item(object):
        def __init__(self, name, status):
            self._name = name
            self._status = status

        @property
        def name(self):
            return self._name

        @property
        def status(self):
            return self._status

    START = Item('start', Status.RUNNING)
    STOP = Item('stop', Status.STOPPED)
    RESTART = Item('restart', Status.RUNNING)



# Copyright (c) 2015, MapR Technologies
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import abc
import os
import re

import jinja2 as j2

import sahara.plugins.exceptions as e
import sahara.plugins.utils as utils
from sahara_plugin_mapr.i18n import _


class FileAttr(object, metaclass=abc.ABCMeta):
    def __init__(self, path, data, mode, owner):
        self.path = path
        self.data = data
        self.mode = mode
        self.owner = owner


class BaseConfigurationFile(object, metaclass=abc.ABCMeta):
    def __init__(self, file_name):
        self.f_name = file_name
        self._config_dict = dict()
        self._local_path = None
        self._remote_path = None
        self.mode = None
        self.owner = None

    @property
    def remote_path(self):
        return self._remote_path

    @remote_path.setter
    def remote_path(self, path):
        self._remote_path = os.path.join(path, self.f_name)

    @abc.abstractmethod
    def render(self):
        pass

    @abc.abstractmethod
    def parse(self, content):
        pass

    def fetch(self, instance):
        with instance.remote() as r:
            content = r.read_file_from(self.remote_path, run_as_root=True)
            self.parse(content)

    def load_properties(self, config_dict):
        for k, v in config_dict.items():
            self.add_property(k, v)

    def add_property(self, name, value):
        self._config_dict[name] = value

    def add_properties(self, properties):
        for prop in properties.items():
            self.add_property(*prop)

    def _get_config_value(self, name):
        return self._config_dict.get(name, None)

    def __repr__(self):
        return '<Configuration file %s>' % self.f_name


class HadoopXML(BaseConfigurationFile):
    def __init__(self, file_name):
        super(HadoopXML, self).__init__(file_name)

    def parse(self, content):
        configs = utils.parse_hadoop_xml_with_name_and_value(content)
        for cfg in configs:
            self.add_property(cfg["name"], cfg["value"])

    def render(self):
        return utils.create_hadoop_xml(self._config_dict)


class RawFile(BaseConfigurationFile):
    def __init__(self, file_name):
        super(RawFile, self).__init__(file_name)

    def render(self):
        return self._config_dict.get('content', '')

    def parse(self, content):
        self._config_dict.update({'content': content})


class PropertiesFile(BaseConfigurationFile):
    def __init__(self, file_name, separator='='):
        super(PropertiesFile, self).__init__(file_name)
        self.separator = separator

    def parse(self, content):
        for line in content.splitlines():
            prop = line.strip()
            if len(prop) == 0:
                continue
            if prop[0] in ['#', '!']:
                continue
            name_value = prop.split(self.separator, 1)
            name = name_value[0]
            # check whether the value is empty
            value = name_value[1] if (len(name_value) == 2) else ''
            self.add_property(name.strip(), value.strip())

    def render(self):
        lines = ['%s%s%s' % (k, self.separator, v) for k, v in
                 self._config_dict.items()]
        return "\n".join(lines) + '\n'


class TemplateFile(BaseConfigurationFile):
    def __init__(self, file_name):
        super(TemplateFile, self).__init__(file_name)
        self._template = None

    @staticmethod
    def _j2_render(template, arg_dict):
        if template:
            return template.render(arg_dict)
        else:
            raise e.PluginsInvalidDataException(
                _('Template object must be defined'))

    def render(self):
        return self._j2_render(self._template, self._config_dict)

    def parse(self, content):
        self._template = j2.Template(content)


class EnvironmentConfig(BaseConfigurationFile):
    def __init__(self, file_name):
        super(EnvironmentConfig, self).__init__(file_name)
        self._lines = []
        self._regex = re.compile(r'export\s+(\w+)=(.+)')
        self._tmpl = 'export %s="%s"'

    def parse(self, content):
        for line in content.splitlines():
            line = self._escape(line)
            match = self._regex.match(line)
            if match:
                name, value = match.groups()
                value = value.replace("\"", '')
                self._lines.append((name, value))
                self.add_property(name, value)
            else:
                self._lines.append(line)

    @staticmethod
    def _escape(string):
        try:
            string = string.decode("utf-8")
        except AttributeError:
            pass
        string = str(string).strip()
        string = string.replace("\"", "")
        return string

    def render(self):
        result = []
        for line in self._lines:
            if isinstance(line, tuple):
                name, value = line
                args = (name, self._config_dict.get(name) or value)
                result.append(self._tmpl % args)
                if name in self._config_dict:
                    del self._config_dict[name]
            else:
                result.append(line)
        extra_ops = [self._tmpl % i for i in self._config_dict.items()]
        return '\n'.join(result + extra_ops) + '\n'



# Copyright (c) 2015, MapR Technologies
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import sahara_plugin_mapr.plugins.mapr.domain.node_process as np
import sahara_plugin_mapr.plugins.mapr.domain.service as s
import sahara_plugin_mapr.plugins.mapr.util.validation_utils as vu


SQOOP_2_SERVER = np.NodeProcess(
    name='sqoop2',
    ui_name='Sqoop2-Server',
    package='mapr-sqoop2-server',
    open_ports=[12000]
)
SQOOP_2_CLIENT = np.NodeProcess(
    name='sqoop-client',
    ui_name='Sqoop2-Client',
    package='mapr-sqoop2-client'
)


class Sqoop2(s.Service):
    def __init__(self):
        super(Sqoop2, self).__init__()
        self._name = 'sqoop'
        self._ui_name = 'Sqoop2'
        self._version = '2.0.0'
        self._node_processes = [SQOOP_2_CLIENT, SQOOP_2_SERVER]
        self._validation_rules = [
            vu.exactly(1, SQOOP_2_SERVER),
        ]

    def post_install(self, cluster_context, instances):
        self._set_service_dir_owner(cluster_context, instances)



# Copyright (c) 2015, MapR Technologies
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from oslo_log import log as logging

import sahara.plugins.utils as u
import sahara_plugin_mapr.plugins.mapr.domain.service as s
import sahara_plugin_mapr.plugins.mapr.services.maprfs.maprfs as maprfs
import sahara_plugin_mapr.plugins.mapr.util.event_log as el

LOG = logging.getLogger(__name__)


class Swift(s.Service):
    HADOOP_SWIFT_JAR = ('plugins/mapr/services/swift/'
                        'resources/hadoop-swift-latest.jar')

    def __init__(self):
        super(Swift, self).__init__()
        self._name = 'swift'
        self._ui_name = 'Swift'
        self._cluster_defaults = ['swift-default.json']

    def install(self, cluster_context, instances):
        # swift does not require any package
        pass

    def configure(self, cluster_context, instances=None):
        instances = instances or cluster_context.get_instances()
        file_servers = cluster_context.filter_instances(instances,
                                                        maprfs.FILE_SERVER)
        self._install_swift_jar(cluster_context, file_servers)

    @el.provision_step("Install Swift service")
    def _install_swift_jar(self, cluster_context, instances):
        LOG.debug('Installing Swift jar')
        jar = u.try_get_file_text(Swift.HADOOP_SWIFT_JAR, 'sahara_plugin_mapr')
        path = '%s/swift.jar' % cluster_context.hadoop_lib

        @el.provision_event()
        def install_on_instance(inst):
            with inst.remote() as r:
                r.write_file_to(path, jar, run_as_root=True)

        for instance in instances:
            install_on_instance(instance)



# Copyright (c) 2015, MapR Technologies
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import collections as c

from oslo_log import log as logging

import sahara.plugins.utils as u
import sahara_plugin_mapr.plugins.mapr.domain.configuration_file as cf
import sahara_plugin_mapr.plugins.mapr.domain.service as s
import sahara_plugin_mapr.plugins.mapr.util.general as g

LOG = logging.getLogger(__name__)

db_spec = c.namedtuple('DatabaseSpec', ['db_name', 'user', 'password'])


class MySQL(s.Service):
    METRICS_SPECS = db_spec('metrics', 'maprmetrics', 'mapr')
    HUE_SPECS = db_spec('hue', 'maprhue', 'mapr')
    METASTORE_SPECS = db_spec('metastore', 'maprmetastore', 'mapr')
    RDBMS_SPECS = db_spec('rdbms', 'maprrdbms', 'mapr')
    OOZIE_SPECS = db_spec('oozie', 'maproozie', 'mapr')
    SENTRY_SPECS = db_spec('sentry', 'maprsentry', 'mapr')

    SELECT_DATA = 'mysql -uroot --skip-column-names -e "%s"| grep -E "\w+"'
    GET_DBS_LIST = SELECT_DATA % 'SHOW DATABASES'
    GET_USERS_HOSTS = (
        SELECT_DATA %
        "SELECT Host FROM mysql.user WHERE mysql.user.User='%s'"  # nosec
    )

    DRIVER_CLASS = 'com.mysql.jdbc.Driver'
    MYSQL_SERVER_PORT = 3306
    MYSQL_INSTALL_SCRIPT = 'plugins/mapr/resources/install_mysql.sh'
    INSTALL_PACKAGES_TIMEOUT = 1800

    def __init__(self):
        super(MySQL, self).__init__()
        self._ui_name = 'MySQL'

    @staticmethod
    def _get_db_daemon_name(distro, distro_version):
        if distro.lower() == 'ubuntu':
            return 'mysql'
        if distro.lower() == 'suse':
            return 'mysqld'
        if distro.lower() in ['centos', 'redhatenterpriseserver', 'redhat']:
            if distro_version.split('.')[0] == '7':
                return 'mariadb'
            return 'mysqld'
        return None

    @staticmethod
    def _execute_script(instance, script_path, script_text=None,
                        user='root', password=None):
        with instance.remote() as r:
            if script_text:
                r.write_file_to(script_path, script_text, run_as_root=True)
            LOG.debug('Executing SQL script {path}'.format(path=script_path))
            r.execute_command(("mysql %s %s < %s" %
                               ('-u' + user if user else '',
                                '-p' + password if password else '',
                                script_path)),
                              run_as_root=True)

    @staticmethod
    def _create_service_db(instance, specs):
        f_name = 'create_db_%s.sql' % specs.db_name
        script = MySQL._create_script_obj(f_name, 'create_database.sql',
                                          db_name=specs.db_name,
                                          user=specs.user,
                                          password=specs.password)
        MySQL._execute_script(instance, script.remote_path, script.render())

    @staticmethod
    def _create_metrics_db(instance, databases, instances):
        if MySQL.METRICS_SPECS.db_name not in databases:
            MySQL._create_service_db(instance, MySQL.METRICS_SPECS)
            MySQL._execute_script(instance=instance,
                                  script_path='/opt/mapr/bin/setup.sql')
        MySQL._grant_access(instance, MySQL.METRICS_SPECS, instances)

    @staticmethod
    def _create_hue_db(instance, databases, instances):
        if MySQL.HUE_SPECS.db_name not in databases:
            MySQL._create_service_db(instance, MySQL.HUE_SPECS)
        MySQL._grant_access(instance, MySQL.HUE_SPECS, instances)

    @staticmethod
    def _create_rdbms_db(instance, databases, instances):
        if MySQL.RDBMS_SPECS.db_name not in databases:
            MySQL._create_service_db(instance, MySQL.RDBMS_SPECS)
        MySQL._grant_access(instance, MySQL.RDBMS_SPECS, instances)

    @staticmethod
    def _create_metastore_db(instance, cluster_context, databases, instances):
        hive_meta = cluster_context.get_instance('HiveMetastore')

        if not hive_meta:
            return

        db_name = MySQL.METASTORE_SPECS.db_name
        if db_name not in databases:
            MySQL._create_service_db(instance, MySQL.METASTORE_SPECS)
        MySQL._grant_access(instance, MySQL.METASTORE_SPECS, instances)

    @staticmethod
    def _create_oozie_db(instance, databases, instances):
        if MySQL.OOZIE_SPECS.db_name not in databases:
            MySQL._create_service_db(instance, MySQL.OOZIE_SPECS)
        MySQL._grant_access(instance, MySQL.OOZIE_SPECS, instances)

    @staticmethod
    def _create_sentry_db(instance, cluster_context, databases, instances):
        sentry_instance = cluster_context.get_instance('Sentry')
        if not sentry_instance:
            return
        if MySQL.SENTRY_SPECS.db_name not in databases:
            MySQL._create_service_db(instance, MySQL.SENTRY_SPECS)
        MySQL._grant_access(instance, MySQL.SENTRY_SPECS, instances)

    @staticmethod
    def start_mysql_server(cluster_context):
        LOG.debug('Starting MySQL Server')
        instance = MySQL.get_db_instance(cluster_context)
        distro = cluster_context.distro
        distro_version = cluster_context.distro_version
        with instance.remote() as r:
            r.execute_command(('service %s restart' %
                               MySQL._get_db_daemon_name(distro.name,
                                                         distro_version)),
                              run_as_root=True)
        LOG.debug('MySQL Server successfully started')

    @staticmethod
    def get_databases_list(db_instance):
        with db_instance.remote() as r:
            ec, out = r.execute_command(MySQL.GET_DBS_LIST)
            if out:
                return out.splitlines()
        return list()

    @staticmethod
    def get_user_hosts(db_instance, username):
        with db_instance.remote() as r:
            ec, out = r.execute_command(MySQL.GET_USERS_HOSTS % username)
            if out:
                return out.splitlines()
        return list()

    @staticmethod
    def get_db_instance(cluster_context):
        return cluster_context.oozie_server

    @staticmethod
    def create_databases(cluster_context, instances):
        db_instance = MySQL.get_db_instance(cluster_context)
        databases = MySQL.get_databases_list(db_instance)
        MySQL._create_metrics_db(db_instance, databases, instances)
        MySQL._create_hue_db(db_instance, databases, instances)
        MySQL._create_rdbms_db(db_instance, databases, instances)
        MySQL._create_oozie_db(db_instance, databases, instances)
        MySQL._create_metastore_db(
            db_instance, cluster_context, databases, instances)
        MySQL._create_sentry_db(db_instance, cluster_context, databases,
                                instances)

    @staticmethod
    def _create_script_obj(filename, template, **kwargs):
        script = cf.TemplateFile(filename)
        script.remote_path = '/tmp/'
        script.parse(u.get_file_text(
            'plugins/mapr/services/mysql/resources/%s' % template,
            'sahara_plugin_mapr'))
        for k, v in kwargs.items():
            script.add_property(k, v)
        return script

    @staticmethod
    def _grant_access(instance, specs, instances):
        f_name = 'grant_access_%s.sql' % specs.db_name
        ips = [i.internal_ip for i in instances]
        user_hosts = MySQL.get_user_hosts(instance, specs.user)
        script = MySQL._create_script_obj(f_name, 'grant_access.sql',
                                          hosts=set(ips) - set(user_hosts),
                                          db_name=specs.db_name,
                                          user=specs.user,
                                          password=specs.password)
        MySQL._execute_script(instance, script.remote_path, script.render())

    @staticmethod
    def install_mysql(instance, distro_name, distro_version):
        g.run_script(instance, MySQL.MYSQL_INSTALL_SCRIPT, 'root', distro_name,
                     distro_version.split('.')[0])



# Copyright (c) 2015, MapR Technologies
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import os

from sahara.plugins import context
from sahara.plugins import edp
from sahara.service.edp.job_binaries import manager as jb_manager
import sahara_plugin_mapr.plugins.mapr.util.maprfs_helper as mfs
import sahara_plugin_mapr.plugins.mapr.versions.version_handler_factory as vhf


class MapROozieJobEngine(edp.PluginsOozieJobEngine):
    def __init__(self, cluster):
        super(MapROozieJobEngine, self).__init__(cluster)
        self.cluster_context = self._get_cluster_context(self.cluster)

    hdfs_user = 'mapr'

    def get_hdfs_user(self):
        return MapROozieJobEngine.hdfs_user

    def create_hdfs_dir(self, remote, dir_name):
        mfs.create_maprfs4_dir(remote, dir_name, self.get_hdfs_user())

    def _upload_workflow_file(self, where, job_dir, wf_xml, hdfs_user):
        f_name = 'workflow.xml'
        with where.remote() as r:
            mfs.put_file_to_maprfs(r, wf_xml, f_name, job_dir, hdfs_user)
        return os.path.join(job_dir, f_name)

    def _upload_job_files_to_hdfs(self, where, job_dir, job, configs,
                                  proxy_configs=None):
        mains = job.mains or []
        libs = job.libs or []
        builtin_libs = edp.get_builtin_binaries(job, configs)
        uploaded_paths = []
        hdfs_user = self.get_hdfs_user()
        lib_dir = job_dir + '/lib'

        with where.remote() as r:
            for m in mains:
                path = jb_manager.JOB_BINARIES. \
                    get_job_binary_by_url(m.url). \
                    copy_binary_to_cluster(m, proxy_configs=proxy_configs,
                                           remote=r, context=context.ctx())
                target = os.path.join(job_dir, m.name)
                mfs.copy_from_local(r, path, target, hdfs_user)
                uploaded_paths.append(target)
            if len(libs) > 0:
                self.create_hdfs_dir(r, lib_dir)
            for l in libs:
                path = jb_manager.JOB_BINARIES. \
                    get_job_binary_by_url(l.url). \
                    copy_binary_to_cluster(l, proxy_configs=proxy_configs,
                                           remote=r, context=context.ctx())
                target = os.path.join(lib_dir, l.name)
                mfs.copy_from_local(r, path, target, hdfs_user)
                uploaded_paths.append(target)
            for lib in builtin_libs:
                mfs.put_file_to_maprfs(r, lib['raw'], lib['name'], lib_dir,
                                       hdfs_user)
                uploaded_paths.append(lib_dir + '/' + lib['name'])
        return uploaded_paths

    def get_name_node_uri(self, cluster):
        return self.cluster_context.name_node_uri

    def get_oozie_server_uri(self, cluster):
        return self.cluster_context.oozie_server_uri

    def get_oozie_server(self, cluster):
        return self.cluster_context.oozie_server

    def get_resource_manager_uri(self, cluster):
        return self.cluster_context.resource_manager_uri

    def _get_cluster_context(self, cluster):
        h_version = cluster.hadoop_version
        v_handler = vhf.VersionHandlerFactory.get().get_handler(h_version)
        return v_handler.get_context(cluster)



# Copyright (c) 2015, MapR Technologies
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import os


def _load_versions():
    d_name = os.path.dirname(__file__)
    m_template = 'sahara_plugin_mapr.plugins.mapr.versions.%s.version_handler'

    def predicate(v_dir):
        if v_dir != "__pycache__":
            v_dir = os.path.join(d_name, v_dir)
            return os.path.isdir(v_dir) and os.listdir(v_dir)

        return False

    def mapper(v_dir):
        return m_template % v_dir

    v_dirs = filter(predicate, os.listdir(d_name))
    m_names = map(mapper, v_dirs)
    versions = {}
    for m_name in m_names:
        m = __import__(m_name, fromlist=['sahara'])
        versions[m.version] = getattr(m, 'VersionHandler')()
    return versions


class VersionHandlerFactory(object):
    instance = None
    versions = None

    @staticmethod
    def get():
        if not VersionHandlerFactory.instance:
            VersionHandlerFactory.versions = _load_versions()
            VersionHandlerFactory.instance = VersionHandlerFactory()
        return VersionHandlerFactory.instance

    def get_versions(self):
        return list(VersionHandlerFactory.versions.keys())

    def get_handler(self, version):
        return VersionHandlerFactory.versions[version]



# Copyright (c) 2015, MapR Technologies
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from sahara.plugins import swift_helper
from sahara.plugins import topology_helper as topo
import sahara_plugin_mapr.plugins.mapr.domain.configuration_file as bcf
import sahara_plugin_mapr.plugins.mapr.domain.node_process as np
import sahara_plugin_mapr.plugins.mapr.domain.service as s
import sahara_plugin_mapr.plugins.mapr.util.validation_utils as vu


RESOURCE_MANAGER = np.NodeProcess(
    name='resourcemanager',
    ui_name='ResourceManager',
    package='mapr-resourcemanager',
    open_ports=[8033, 8032, 8031, 8030, 8088]
)
NODE_MANAGER = np.NodeProcess(
    name='nodemanager',
    ui_name='NodeManager',
    package='mapr-nodemanager',
    open_ports=[8041, 8040, 8042, 8044]
)
HISTORY_SERVER = np.NodeProcess(
    name='historyserver',
    ui_name='HistoryServer',
    package='mapr-historyserver',
    open_ports=[10020, 19888, 19890]
)


class YARN(s.Service):
    cluster_mode = 'yarn'

    def __init__(self):
        super(YARN, self).__init__()
        self._name = 'hadoop'
        self._ui_name = 'YARN'
        self._node_processes = [RESOURCE_MANAGER, NODE_MANAGER, HISTORY_SERVER]
        self._ui_info = [
            ('NodeManager', NODE_MANAGER, {s.SERVICE_UI: 'http://%s:8042'}),
            ('ResourceManager', RESOURCE_MANAGER,
             {s.SERVICE_UI: 'http://%s:8088'}),
            ('HistoryServer', RESOURCE_MANAGER,
             {s.SERVICE_UI: 'http://%s:19888'}),
        ]
        self._cluster_defaults = ['yarn-cluster.json']
        self._node_defaults = ['yarn-node.json']

    def get_config_files(self, cluster_context, configs, instance=None):
        # yarn-site.xml
        yarn_site = bcf.HadoopXML("yarn-site.xml")
        yarn_site.remote_path = self.conf_dir(cluster_context)
        if instance:
            yarn_site.fetch(instance)
        yarn_site.add_properties(self._get_yarn_site_props(cluster_context))
        yarn_site.load_properties(configs)

        # core-site.xml
        core_site = bcf.HadoopXML("core-site.xml")
        core_site.remote_path = self.conf_dir(cluster_context)
        if instance:
            core_site.fetch(instance)
        core_site.add_properties(self._get_core_site_props(cluster_context))

        return [yarn_site, core_site]

    def _get_core_site_props(self, cluster_context):
        result = {
            'hadoop.proxyuser.mapr.groups': '*',
            'hadoop.proxyuser.mapr.hosts': '*',
        }
        if cluster_context.is_node_aware:
            result.update(self._get_core_site_node_aware_props())
        for conf in swift_helper.get_swift_configs():
            result[conf['name']] = conf['value']
        return result

    def _get_core_site_node_aware_props(self):
        result = topo.vm_awareness_core_config()
        result = {c['name']: c['value'] for c in result}
        result.update({
            'net.topology.script.file.name': '/opt/mapr/topology.sh',
            'net.topology.script.number.args': '75',
        })
        return result

    def _get_yarn_site_props(self, cluster_context):
        return {
            'hadoop.proxyuser.mapr.groups': '*',
            'hadoop.proxyuser.mapr.hosts': '*',
        }

    def conf_dir(self, cluster_context):
        return '%s/etc/hadoop' % self.home_dir(cluster_context)

    def get_file_path(self, file_name):
        template = 'plugins/mapr/services/yarn/resources/%s'
        return template % file_name


class YARNv241(YARN):
    def __init__(self):
        super(YARNv241, self).__init__()
        self._version = '2.4.1'
        self._validation_rules = [
            vu.exactly(1, RESOURCE_MANAGER),
            vu.at_least(1, NODE_MANAGER),
            vu.exactly(1, HISTORY_SERVER),
        ]


class YARNv251(YARN):
    def __init__(self):
        super(YARNv251, self).__init__()
        self._version = '2.5.1'
        self._validation_rules = [
            vu.at_least(1, RESOURCE_MANAGER),
            vu.at_least(1, NODE_MANAGER),
            vu.exactly(1, HISTORY_SERVER),
        ]


class YARNv270(YARN):
    def __init__(self):
        super(YARNv270, self).__init__()
        self._version = "2.7.0"
        self._validation_rules = [
            vu.at_least(1, RESOURCE_MANAGER),
            vu.at_least(1, NODE_MANAGER),
            vu.exactly(1, HISTORY_SERVER),
        ]



# Copyright (c) 2015, MapR Technologies
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


def chown(instance, owner, path, run_as_root=True):
    cmd = 'chown -R %(owner)s %(path)s' % {'owner': owner, 'path': path}
    with instance.remote() as r:
        r.execute_command(cmd, run_as_root=run_as_root)


def re_configure_sh(instance, cluster_context):
    with instance.remote() as r:
        command = '%s -R' % cluster_context.configure_sh_path
        r.execute_command(command, run_as_root=True)



# Copyright (c) 2015, MapR Technologies
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from oslo_log import log as logging

from sahara.i18n import _
import sahara.utils.files as files
import sahara_plugin_mapr.plugins.mapr.domain.configuration_file as bcf
import sahara_plugin_mapr.plugins.mapr.domain.node_process as np
import sahara_plugin_mapr.plugins.mapr.domain.service as s
import sahara_plugin_mapr.plugins.mapr.services.sentry.sentry as sentry
import sahara_plugin_mapr.plugins.mapr.util.event_log as el
import sahara_plugin_mapr.plugins.mapr.util.validation_utils as vu

LOG = logging.getLogger(__name__)

HIVE_METASTORE = np.NodeProcess(
    name='hivemeta',
    ui_name='HiveMetastore',
    package='mapr-hivemetastore',
    open_ports=[9083]
)
HIVE_SERVER_2 = np.NodeProcess(
    name='hs2',
    ui_name='HiveServer2',
    package='mapr-hiveserver2',
    open_ports=[10000]
)


class Hive(s.Service):
    def __init__(self):
        super(Hive, self).__init__()
        self._name = 'hive'
        self._ui_name = 'Hive'
        self._node_processes = [HIVE_METASTORE, HIVE_SERVER_2]
        self._validation_rules = [
            vu.exactly(1, HIVE_METASTORE),
            vu.exactly(1, HIVE_SERVER_2),
        ]

    # hive-site.xml
    def get_config_files(self, cluster_context, configs, instance=None):
        hive_default = 'plugins/mapr/services/hive/resources/hive-default.xml'
        hive_site = bcf.HadoopXML("hive-site.xml")
        hive_site.remote_path = self.conf_dir(cluster_context)
        if instance:
            hive_site.fetch(instance)
        hive_site.parse(files.get_file_text(hive_default,
                                            'sahara_plugin_mapr'))
        hive_site.add_properties(self._get_hive_site_props(cluster_context))
        sentry_host = cluster_context.get_instance(sentry.SENTRY)
        if sentry_host:
            sentry_mode = cluster_context._get_cluster_config_value(
                sentry.Sentry().SENTRY_STORAGE_MODE)
            ui_name = sentry.Sentry().ui_name
            sentry_version = cluster_context.get_chosen_service_version(
                ui_name)
            sentry_service = cluster_context. \
                _find_service_instance(ui_name, sentry_version)
            if sentry_service.supports(self, sentry_mode):
                sentry_default = 'plugins/mapr/services/hive/resources/' \
                                 'sentry-default.xml'
                sentry_db = \
                    'plugins/mapr/services/hive/resources/sentry-db.xml'
                hive_site.parse(files.get_file_text(sentry_default,
                                                    'sahara_plugin_mapr'))
                hive_site.add_property('hive.sentry.conf.url',
                                       'file://%s/sentry-site.xml' %
                                       sentry_service.conf_dir(
                                           cluster_context))
                if sentry_mode == sentry.DB_STORAGE_SENTRY_MODE:
                    hive_site.parse(files.get_file_text(sentry_db,
                                                        'sahara_plugin_mapr'))

        return [hive_site]

    def _get_hive_site_props(self, cluster_context):
        # Import here to resolve circular dependency
        from sahara_plugin_mapr.plugins.mapr.services.mysql import mysql

        zookeepers = cluster_context.get_zookeeper_nodes_ip()
        metastore_specs = mysql.MySQL.METASTORE_SPECS

        return {
            'javax.jdo.option.ConnectionDriverName': mysql.MySQL.DRIVER_CLASS,
            'javax.jdo.option.ConnectionURL': self._get_jdbc_uri(
                cluster_context),
            'javax.jdo.option.ConnectionUserName': metastore_specs.user,
            'javax.jdo.option.ConnectionPassword': metastore_specs.password,
            'hive.metastore.uris': self._get_metastore_uri(cluster_context),
            'hive.zookeeper.quorum': zookeepers,
            'hbase.zookeeper.quorum': zookeepers,
        }

    def _get_jdbc_uri(self, cluster_context):
        # Import here to resolve circular dependency
        from sahara_plugin_mapr.plugins.mapr.services.mysql import mysql

        jdbc_uri = ('jdbc:mysql://%(db_host)s:%(db_port)s/%(db_name)s?'
                    'createDatabaseIfNotExist=true')
        jdbc_args = {
            'db_host': mysql.MySQL.get_db_instance(
                cluster_context).internal_ip,
            'db_port': mysql.MySQL.MYSQL_SERVER_PORT,
            'db_name': mysql.MySQL.METASTORE_SPECS.db_name,
        }
        return jdbc_uri % jdbc_args

    def _get_metastore_uri(self, cluster_context):
        return 'thrift://%s:9083' % cluster_context.get_instance_ip(
            HIVE_METASTORE)

    def post_start(self, cluster_context, instances):
        # Import here to resolve circular dependency
        import sahara_plugin_mapr.plugins.mapr.services.maprfs.maprfs as mfs

        create_path = lambda p: 'sudo -u mapr hadoop fs -mkdir %s' % p
        check_path = 'sudo -u mapr hadoop fs -ls %s'
        cmd = "%(check)s || ( %(parent)s && %(target)s )"
        args = {
            'check': check_path % '/user/hive/warehouse/',
            'parent': create_path('/user/hive/'),
            'target': create_path('/user/hive/warehouse/')
        }
        cldb_node = cluster_context.get_instance(mfs.CLDB)
        with cldb_node.remote() as r:
            LOG.debug("Creating Hive warehouse dir")
            r.execute_command(cmd % args, raise_when_error=False)
        self._create_sentry_role(cluster_context)

    def _create_sentry_role(self, cluster_context):

        @el.provision_event(name=_("Create Sentry role for Hive"))
        def _create_role(instance):
            cmd = 'sudo -u mapr hive -e "create role admin_role;' \
                  'grant all on server HS2 to role admin_role;' \
                  'grant role admin_role to group mapr;"'
            with instance.remote() as r:
                LOG.debug("Creating hive role for sentry")
                r.execute_command(cmd, raise_when_error=False)

        hive_host = cluster_context.get_instance(HIVE_METASTORE)
        sentry_host = cluster_context.get_instance(sentry.SENTRY)
        if sentry_host:
            sentry_mode = cluster_context._get_cluster_config_value(
                sentry.Sentry().SENTRY_STORAGE_MODE)
            ui_name = sentry.Sentry().ui_name
            sentry_version = cluster_context.get_chosen_service_version(
                ui_name)
            sentry_service = cluster_context. \
                _find_service_instance(ui_name, sentry_version)
            if sentry_service.supports(self, sentry_mode):
                _create_role(hive_host)


class HiveV013(Hive):
    def __init__(self):
        super(HiveV013, self).__init__()
        self._version = '0.13'
        self._dependencies = [('mapr-hive', self.version)]


class HiveV10(Hive):
    def __init__(self):
        super(HiveV10, self).__init__()
        self._version = "1.0"
        self._dependencies = [("mapr-hive", self.version)]


class HiveV12(Hive):
    def __init__(self):
        super(HiveV12, self).__init__()
        self._version = "1.2"
        self._dependencies = [("mapr-hive", self.version)]



# Copyright (c) 2015, MapR Technologies
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import sahara_plugin_mapr.plugins.mapr.domain.configuration_file as conf_f
import sahara_plugin_mapr.tests.unit.base as b


class TestHadoopXML(b.SaharaTestCase):
    def __init__(self, *args, **kwds):
        super(TestHadoopXML, self).__init__(*args, **kwds)
        self.content = '''<?xml version="1.0"?>
<?xml-stylesheet type="text/xsl" href="configuration.xsl"?>
<configuration>
    <property>
        <name>key1</name>
        <value>value1</value>
    </property>
    <property>
        <name>key2</name>
        <value>value2</value>
    </property>
</configuration>'''

    def test_remote_path(self):
        foo = conf_f.HadoopXML('foo')
        foo.remote_path = '/bar'
        self.assertEqual('/bar/foo', foo.remote_path)

    def test_parse(self):
        foo = conf_f.HadoopXML('foo')
        foo.parse(self.content)
        expected = {'key1': 'value1', 'key2': 'value2'}
        self.assertEqual(expected, foo._config_dict)

    def test_render(self):
        foo = conf_f.HadoopXML('foo')
        expected = {'ke1': 'value1', 'key2': 'value2'}
        foo._config_dict = expected
        actual = foo.render()
        bar = conf_f.HadoopXML('bar')
        bar.parse(actual)
        self.assertEqual(expected, bar._config_dict)

    def test_add_property(self):
        foo = conf_f.HadoopXML('foo')
        self.assertEqual(foo._config_dict, {})
        foo.add_property('key1', 'value1')
        self.assertEqual(foo._config_dict, {'key1': 'value1'})
        foo.add_property('key2', 'value2')
        expected = {'key1': 'value1', 'key2': 'value2'}
        self.assertEqual(expected, foo._config_dict)

    def test_get_config_value(self):
        foo = conf_f.HadoopXML('foo')
        foo._config_dict = {'foo': 'bar'}
        self.assertEqual('bar', foo._get_config_value('foo'))
        self.assertIsNone(foo._get_config_value('bar'))


class TestRawFile(b.SaharaTestCase):
    def __init__(self, *args, **kwds):
        super(TestRawFile, self).__init__(*args, **kwds)
        self.content = 'some meaningful text'

    def test_remote_path(self):
        foo = conf_f.RawFile('foo')
        foo.remote_path = '/bar'
        self.assertEqual('/bar/foo', foo.remote_path)

    def test_parse(self):
        foo = conf_f.RawFile('foo')
        foo.parse(self.content)
        expected = {'content': self.content}
        self.assertEqual(expected, foo._config_dict)

    def test_render(self):
        foo = conf_f.RawFile('foo')
        expected = {'content': 'foo bar'}
        foo._config_dict = expected
        actual = foo.render()
        bar = conf_f.RawFile('bar')
        bar.parse(actual)
        self.assertEqual(expected, bar._config_dict)


class TestPropertiesFile(b.SaharaTestCase):
    def __init__(self, *args, **kwds):
        super(TestPropertiesFile, self).__init__(*args, **kwds)
        self.content = '''
key1=value1
key2=value2
'''

    def test_remote_path(self):
        foo = conf_f.PropertiesFile('foo')
        foo.remote_path = '/bar'
        self.assertEqual('/bar/foo', foo.remote_path)

    def test_parse(self):
        foo = conf_f.PropertiesFile('foo')
        foo.parse(self.content)
        expected = {'key1': 'value1', 'key2': 'value2'}
        self.assertEqual(expected, foo._config_dict)

    def test_render(self):
        foo = conf_f.PropertiesFile('foo')
        expected = {'ke1': 'value1', 'key2': 'value2'}
        foo._config_dict = expected
        actual = foo.render()
        bar = conf_f.PropertiesFile('bar')
        bar.parse(actual)
        self.assertEqual(expected, bar._config_dict)

    def test_add_property(self):
        foo = conf_f.PropertiesFile('foo')
        expected = {}
        self.assertEqual(expected, foo._config_dict)

        foo.add_property('key1', 'value1')
        expected = {'key1': 'value1'}
        self.assertEqual(expected, foo._config_dict)

        foo.add_property('key2', 'value2')
        expected = {'key1': 'value1', 'key2': 'value2'}
        self.assertEqual(expected, foo._config_dict)

    def test_get_config_value(self):
        foo = conf_f.PropertiesFile('foo')
        foo._config_dict = {'foo': 'bar'}
        self.assertEqual('bar', foo._get_config_value('foo'))
        self.assertIsNone(foo._get_config_value('bar'))


class TestTemplateFile(b.SaharaTestCase):
    def __init__(self, *args, **kwds):
        super(TestTemplateFile, self).__init__(*args, **kwds)
        self.content = '''
key1={{ value1 }}
key2={{ value2 }}'''
        self.rendered = '''
key1=value1
key2=value2'''

    def test_remote_path(self):
        foo = conf_f.TemplateFile('foo')
        foo.remote_path = '/bar'
        self.assertEqual('/bar/foo', foo.remote_path)

    def test_parse(self):
        foo = conf_f.TemplateFile('foo')
        foo.parse(self.content)
        self.assertIsNotNone(foo._template)

    def test_render(self):
        foo = conf_f.TemplateFile('foo')
        expected = {'value1': 'value1', 'value2': 'value2'}
        foo.parse(self.content)
        foo._config_dict = expected
        actual = foo.render()
        self.assertEqual(self.rendered, actual)

    def test_add_property(self):
        foo = conf_f.TemplateFile('foo')
        expected = {}
        self.assertEqual(expected, foo._config_dict)

        foo.add_property('key1', 'value1')
        expected = {'key1': 'value1'}
        self.assertEqual(expected, foo._config_dict)

        foo.add_property('key2', 'value2')
        expected = {'key1': 'value1', 'key2': 'value2'}
        self.assertEqual(expected, foo._config_dict)


class TestEnvironmentConfig(b.SaharaTestCase):
    def __init__(self, *args, **kwds):
        super(TestEnvironmentConfig, self).__init__(*args, **kwds)
        self.content = '''
non export line
export key1=value1
export key2=value2
export key
'''

    def test_remote_path(self):
        foo = conf_f.EnvironmentConfig('foo')
        foo.remote_path = '/bar'
        self.assertEqual('/bar/foo', foo.remote_path)

    def test_parse(self):
        foo = conf_f.EnvironmentConfig('foo')
        foo.parse(self.content)
        expected = {'key1': 'value1', 'key2': 'value2'}
        self.assertEqual(expected, foo._config_dict)

    def test_render(self):
        foo = conf_f.EnvironmentConfig('foo')
        expected = {'ke1': 'value1', 'key2': 'value2'}
        foo._config_dict = expected
        actual = foo.render()
        bar = conf_f.EnvironmentConfig('bar')
        bar.parse(actual)
        self.assertEqual(expected, bar._config_dict)

    def test_render_extra_properties(self):
        foo = conf_f.EnvironmentConfig('foo')
        foo.parse(self.content)
        foo.add_property('key3', 'value3')
        foo_content = foo.render()
        bar = conf_f.EnvironmentConfig('bar')
        bar.parse(foo_content)
        expected = {'key1': 'value1', 'key2': 'value2', 'key3': 'value3'}
        self.assertEqual(expected, bar._config_dict)

    def test_add_property(self):
        foo = conf_f.EnvironmentConfig('foo')
        self.assertEqual({}, foo._config_dict)
        foo.add_property('key1', 'value1')
        self.assertEqual({'key1': 'value1'}, foo._config_dict)
        foo.add_property('key2', 'value2')
        expected = {'key1': 'value1', 'key2': 'value2'}
        self.assertEqual(expected, foo._config_dict)

    def test_get_config_value(self):
        foo = conf_f.EnvironmentConfig('foo')
        foo._config_dict = {'foo': 'bar'}
        self.assertEqual('bar', foo._get_config_value('foo'))
        self.assertIsNone(foo._get_config_value('bar'))



# Copyright (c) 2015, MapR Technologies
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from oslo_log import log as logging
from oslo_serialization import jsonutils as json

import sahara.plugins.exceptions as ex
import sahara.plugins.provisioning as p
from sahara.plugins import utils
from sahara_plugin_mapr.i18n import _
from sahara_plugin_mapr.plugins.mapr.util import commands as cmd
from sahara_plugin_mapr.plugins.mapr.util import event_log as el
from sahara_plugin_mapr.plugins.mapr.util import general as g
from sahara_plugin_mapr.plugins.mapr.util import service_utils as su

LOG = logging.getLogger(__name__)

SERVICE_UI = 'Web UI'

_INSTALL_PACKAGES_TIMEOUT = 3600


class Service(object, metaclass=g.Singleton):
    def __init__(self):
        self._name = None
        self._ui_name = None
        self._node_processes = []
        self._version = None
        self._dependencies = []
        self._ui_info = []
        self._cluster_defaults = []
        self._node_defaults = []
        self._validation_rules = []
        self._priority = 1

    @property
    def name(self):
        return self._name

    @property
    def ui_name(self):
        return self._ui_name

    @property
    def version(self):
        return self._version

    @property
    def node_processes(self):
        return self._node_processes

    @property
    def dependencies(self):
        return self._dependencies

    @property
    def cluster_defaults(self):
        return self._cluster_defaults

    @property
    def node_defaults(self):
        return self._node_defaults

    @property
    def validation_rules(self):
        return self._validation_rules

    def get_ui_info(self, cluster_context):
        return self._ui_info

    def install(self, cluster_context, instances):
        service_instances = cluster_context.filter_instances(instances,
                                                             service=self)

        @el.provision_step(_("Install %s service") % self.ui_name,
                           cluster_context_reference=0, instances_reference=1)
        def _install(_context, _instances):
            g.execute_on_instances(_instances,
                                   self._install_packages_on_instance,
                                   _context)
        if service_instances:
            _install(cluster_context, service_instances)

    @el.provision_event(instance_reference=1)
    def _install_packages_on_instance(self, instance, cluster_context):
        processes = [p for p in self.node_processes if
                     p.ui_name in instance.node_group.node_processes]
        if processes is not None and len(processes) > 0:
            packages = self._get_packages(cluster_context, processes)
            cmd = cluster_context.distro.create_install_cmd(packages)
            with instance.remote() as r:
                r.execute_command(cmd, run_as_root=True,
                                  timeout=_INSTALL_PACKAGES_TIMEOUT)

    def _get_packages(self, cluster_context, node_processes):
        result = []

        result += self.dependencies
        result += [(np.package, self.version) for np in node_processes]

        return result

    def _set_service_dir_owner(self, cluster_context, instances):
        service_instances = cluster_context.filter_instances(instances,
                                                             service=self)
        LOG.debug("Changing %s service dir owner", self.ui_name)
        for instance in service_instances:
            cmd.chown(instance, 'mapr:mapr', self.service_dir(cluster_context))

    def post_install(self, cluster_context, instances):
        pass

    def post_start(self, cluster_context, instances):
        pass

    def configure(self, cluster_context, instances=None):
        pass

    def update(self, cluster_context, instances=None):
        pass

    def get_file_path(self, file_name):
        template = 'plugins/mapr/services/%(service)s/resources/%(file_name)s'
        args = {'service': self.name, 'file_name': file_name}
        return template % args

    def get_configs(self):
        result = []

        for d_file in self.cluster_defaults:
            data = self._load_config_file(self.get_file_path(d_file))
            result += [self._create_config_obj(c, self.ui_name) for c in data]

        for d_file in self.node_defaults:
            data = self._load_config_file(self.get_file_path(d_file))
            result += [self._create_config_obj(c, self.ui_name, scope='node')
                       for c in data]

        return result

    def get_configs_dict(self):
        result = dict()
        for conf_obj in self.get_configs():
            result.update({conf_obj.name: conf_obj.default_value})
        return {self.ui_name: result}

    def _load_config_file(self, file_path=None):
        return json.loads(utils.get_file_text(file_path, 'sahara_plugin_mapr'))

    def get_config_files(self, cluster_context, configs, instance=None):
        return []

    def _create_config_obj(self, item, target='general', scope='cluster',
                           high_priority=False):
        def _prepare_value(value):
            if isinstance(value, str):
                return value.strip().lower()
            return value

        conf_name = _prepare_value(item.get('name', None))

        conf_value = _prepare_value(item.get('value', None))

        if not conf_name:
            raise ex.HadoopProvisionError(_("Config missing 'name'"))

        if conf_value is None:
            raise ex.PluginInvalidDataException(
                _("Config '%s' missing 'value'") % conf_name)

        if high_priority or item.get('priority', 2) == 1:
            priority = 1
        else:
            priority = 2

        return p.Config(
            name=conf_name,
            applicable_target=target,
            scope=scope,
            config_type=item.get('config_type', "string"),
            config_values=item.get('config_values', None),
            default_value=conf_value,
            is_optional=item.get('is_optional', True),
            description=item.get('description', None),
            priority=priority)

    def get_version_config(self, versions):
        return p.Config(
            name='%s Version' % self._ui_name,
            applicable_target=self.ui_name,
            scope='cluster',
            config_type='dropdown',
            config_values=[(v, v) for v in sorted(versions, reverse=True)],
            is_optional=False,
            description=_('Specify the version of the service'),
            priority=1)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            version_eq = self.version == other.version
            ui_name_eq = self.ui_name == other.ui_name
            return version_eq and ui_name_eq
        return NotImplemented

    def restart(self, instances):
        for node_process in self.node_processes:
            filtered_instances = su.filter_by_node_process(instances,
                                                           node_process)
            if filtered_instances:
                node_process.restart(filtered_instances)

    def service_dir(self, cluster_context):
        args = {'mapr_home': cluster_context.mapr_home, 'name': self.name}
        return '%(mapr_home)s/%(name)s' % args

    def home_dir(self, cluster_context):
        args = {
            'service_dir': self.service_dir(cluster_context),
            'name': self.name,
            'version': self.version,
        }
        return '%(service_dir)s/%(name)s-%(version)s' % args

    def conf_dir(self, cluster_context):
        return '%s/conf' % self.home_dir(cluster_context)

    def post_configure_sh(self, cluster_context, instances):
        pass

    def post_configure(self, cluster_context, instances):
        pass



# Copyright (c) 2015, MapR Technologies
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import sahara.plugins.provisioning as p
from sahara_plugin_mapr.i18n import _
import sahara_plugin_mapr.plugins.mapr.versions.version_handler_factory as vhf


class MapRPlugin(p.ProvisioningPluginBase):
    title = 'MapR Hadoop Distribution'
    description = _('The MapR Distribution provides a full Hadoop stack that'
                    ' includes the MapR File System (MapR-FS), MapReduce,'
                    ' a complete Hadoop ecosystem, and the MapR Control System'
                    ' user interface')

    def _get_handler(self, hadoop_version):
        return vhf.VersionHandlerFactory.get().get_handler(hadoop_version)

    def get_title(self):
        return MapRPlugin.title

    def get_description(self):
        return MapRPlugin.description

    def get_labels(self):
        return {
            'plugin_labels': {'enabled': {'status': True}},
            'version_labels': {
                '5.2.0.mrv2': {'enabled': {'status': True}},
            }
        }

    def get_versions(self):
        return vhf.VersionHandlerFactory.get().get_versions()

    def get_node_processes(self, hadoop_version):
        return self._get_handler(hadoop_version).get_node_processes()

    def get_configs(self, hadoop_version):
        return self._get_handler(hadoop_version).get_configs()

    def configure_cluster(self, cluster):
        self._get_handler(cluster.hadoop_version).configure_cluster(cluster)

    def start_cluster(self, cluster):
        self._get_handler(cluster.hadoop_version).start_cluster(cluster)

    def validate(self, cluster):
        self._get_handler(cluster.hadoop_version).validate(cluster)

    def validate_scaling(self, cluster, existing, additional):
        v_handler = self._get_handler(cluster.hadoop_version)
        v_handler.validate_scaling(cluster, existing, additional)

    def scale_cluster(self, cluster, instances):
        v_handler = self._get_handler(cluster.hadoop_version)
        v_handler.scale_cluster(cluster, instances)

    def decommission_nodes(self, cluster, instances):
        v_handler = self._get_handler(cluster.hadoop_version)
        v_handler.decommission_nodes(cluster, instances)

    def get_edp_engine(self, cluster, job_type):
        v_handler = self._get_handler(cluster.hadoop_version)
        return v_handler.get_edp_engine(cluster, job_type)

    def get_edp_job_types(self, versions=None):
        res = {}
        for vers in self.get_versions():
            if not versions or vers in versions:
                vh = self._get_handler(vers)
                res[vers] = vh.get_edp_job_types()
        return res

    def get_edp_config_hints(self, job_type, version):
        v_handler = self._get_handler(version)
        return v_handler.get_edp_config_hints(job_type)

    def get_open_ports(self, node_group):
        v_handler = self._get_handler(node_group.cluster.hadoop_version)
        return v_handler.get_open_ports(node_group)

    def get_health_checks(self, cluster):
        v_handler = self._get_handler(cluster.hadoop_version)
        return v_handler.get_cluster_checks(cluster)

    def get_image_arguments(self, hadoop_version):
        return self._get_handler(hadoop_version).get_image_arguments()

    def pack_image(self, hadoop_version, remote,
                   test_only=False, image_arguments=None):
        version = self._get_handler(hadoop_version)
        version.pack_image(hadoop_version, remote, test_only=test_only,
                           image_arguments=image_arguments)

    def validate_images(self, cluster, test_only=False, image_arguments=None):
        self._get_handler(cluster.hadoop_version).validate_images(
            cluster, test_only=test_only, image_arguments=image_arguments)



# Copyright (c) 2015, MapR Technologies
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import sahara.plugins.utils as utils
import sahara_plugin_mapr.plugins.mapr.domain.configuration_file as bcf
import sahara_plugin_mapr.plugins.mapr.domain.node_process as np
import sahara_plugin_mapr.plugins.mapr.domain.service as s
import sahara_plugin_mapr.plugins.mapr.services.hbase.hbase as hbase
import sahara_plugin_mapr.plugins.mapr.services.hive.hive as hive
import sahara_plugin_mapr.plugins.mapr.util.general as g
import sahara_plugin_mapr.plugins.mapr.util.maprfs_helper as mfs
import sahara_plugin_mapr.plugins.mapr.util.validation_utils as vu

SPARK_SLAVE_UI_PORT = 8081
SPARK_HS_UI_PORT = 18080

SPARK_HISTORY_SERVER = np.NodeProcess(
    name='spark-historyserver',
    ui_name='Spark HistoryServer',
    package='mapr-spark-historyserver',
    open_ports=[SPARK_HS_UI_PORT]
)
SPARK_SLAVE = np.NodeProcess(
    name='spark-master',
    ui_name='Spark Slave',
    package='mapr-spark',
    open_ports=[SPARK_SLAVE_UI_PORT]
)


class SparkOnYarn(s.Service):
    JAR_FILE_TARGET = '/apps/spark/lib'
    MFS_DIR = '/apps/spark'
    SERVLET_JAR = 'javax.servlet-api.jar'

    def __init__(self):
        super(SparkOnYarn, self).__init__()
        self._name = 'spark'
        self._ui_name = 'Spark'
        self._version = '1.5.2'
        self._dependencies = [('mapr-spark', self.version)]
        self._node_processes = [
            SPARK_HISTORY_SERVER,
            SPARK_SLAVE,
        ]
        self._validation_rules = [
            vu.exactly(1, SPARK_HISTORY_SERVER),
            vu.at_least(1, SPARK_SLAVE),
        ]
        self._ui_info = [
            ('Spark History Server', SPARK_HISTORY_SERVER,
             {s.SERVICE_UI: 'http://%%s:%s' % SPARK_HS_UI_PORT})]

    def _get_packages(self, cluster_context, node_processes):
        result = []

        result += self.dependencies
        result += [(np.package, self.version) for np in node_processes]
        hbase_version = self._get_hbase_version(cluster_context)
        hive_version = self._get_hive_version(cluster_context)
        if hbase_version:
            result += [('mapr-hbase', hbase_version)]
        if hive_version:
            result += [('mapr-hive', hive_version)]

        return result

    def get_config_files(self, cluster_context, configs, instance=None):
        hbase_version = self._get_hbase_version(cluster_context)
        hive_version = self._get_hive_version(cluster_context)
        # spark-env-sh
        template = 'plugins/mapr/services/' \
                   'spark/resources/spark-env.template'
        env_sh = bcf.TemplateFile('spark-env.sh')
        env_sh.remote_path = self.conf_dir(cluster_context)
        env_sh.parse(utils.get_file_text(template, 'sahara_plugin_mapr'))
        env_sh.add_property('version', self.version)
        env_sh.add_property('servlet_api_jar', self.SERVLET_JAR)

        # spark-defaults
        conf = bcf.PropertiesFile('spark-defaults.conf', separator=' ')
        conf.remote_path = self.conf_dir(cluster_context)
        if instance:
            conf.fetch(instance)
        conf.add_property('spark.yarn.jar', 'maprfs://%s/%s' %
                          (self.JAR_FILE_TARGET,
                           self._assembly_jar_path(cluster_context)
                           .rsplit('/', 1)[1]))

        # compatibility.version
        versions = bcf.PropertiesFile('compatibility.version')
        versions.remote_path = self.home_dir(cluster_context) + '/mapr-util'
        if instance:
            versions.fetch(instance)

        if hive_version:
            versions.add_property('hive_versions',
                                  hive_version + '.0')
            conf.add_properties(self._hive_properties(cluster_context))
        if hbase_version:
            versions.add_property('hbase_versions', hbase_version)
            conf.add_property('spark.executor.extraClassPath',
                              '%s/lib/*' % self._hbase(cluster_context)
                              .home_dir(cluster_context))
        return [conf, versions, env_sh]

    def update(self, cluster_context, instances=None):
        pass

    def post_install(self, cluster_context, instances):
        pass

    def configure(self, cluster_context, instances=None):
        pass

    def post_start(self, cluster_context, instances):
        self._create_hadoop_spark_dirs(cluster_context)
        self.copy_assembly_jar_to_mfs(cluster_context)
        self._copy_hive_site(cluster_context)
        self._copy_hbase_site(cluster_context)
        self._copy_jar_from_hue(cluster_context)

    def copy_assembly_jar_to_mfs(self, cluster_context):
        target = self.JAR_FILE_TARGET
        hdfs_user = 'mapr'
        with cluster_context.get_instance(SPARK_HISTORY_SERVER).remote() as r:
            mfs.copy_from_local(r, self._assembly_jar_path(cluster_context),
                                target, hdfs_user)

    def _copy_hive_site(self, cluster_context):
        if not self._hive(cluster_context):
            return
        hive_site = self._hive(cluster_context).conf_dir(
            cluster_context) + '/hive-site.xml'
        path = self.conf_dir(cluster_context) + '/hive-site.xml'
        hive_instance = cluster_context.get_instance(hive.HIVE_SERVER_2)
        spark_instance = cluster_context.get_instance(
            SPARK_HISTORY_SERVER)
        g.copy_file(hive_site, hive_instance, path, spark_instance,
                    run_as='root', owner='mapr')

    def _copy_hbase_site(self, cluster_context):
        if not self._hbase(cluster_context):
            return
        hbase_site = self._hbase(cluster_context).conf_dir(
            cluster_context) + '/hbase-site.xml'
        path = self.conf_dir(cluster_context) + '/hbase-site.xml'
        hbase_instance = cluster_context.get_instance(hbase.HBASE_MASTER)
        spark_instance = cluster_context.get_instance(
            SPARK_HISTORY_SERVER)
        g.copy_file(hbase_site, hbase_instance, path, spark_instance,
                    run_as='root', owner='mapr')

    def _create_hadoop_spark_dirs(self, cluster_context):
        home = '/apps/spark'
        libs = self.JAR_FILE_TARGET
        run_as_user = 'mapr'
        with cluster_context.get_instance(SPARK_HISTORY_SERVER).remote() as r:
            mfs.mkdir(r, home, run_as=run_as_user)
            mfs.mkdir(r, libs, run_as=run_as_user)
            mfs.chmod(r, home, 777, run_as=run_as_user)
            mfs.chmod(r, libs, 777, run_as=run_as_user)

    def _hive_properties(self, cluster_context):
        hive_version = self._hive(cluster_context).version
        hive_conf = self._hive(cluster_context).conf_dir(cluster_context)
        hive_site = hive_conf + '/hive-site.xml'
        hive_datanucleus_libs = self._hive_datanucleus_libs_path(
            cluster_context)
        hive_libs = self._hive_libs_path(cluster_context)
        hadoop_libs = self._hadoop_libs_path(cluster_context)
        hive_datanucleus_libs.insert(0, hive_site)
        mfs_paths = self._hive_datanucleus_libs_path(cluster_context)
        return {
            'spark.yarn.dist.files': ','.join(mfs_paths),
            'spark.sql.hive.metastore.version': hive_version + '.0',
            'spark.sql.hive.metastore.jars': ':'.join(hadoop_libs + hive_libs)
        }

    def _hadoop_libs_path(self, cluster_context):
        cmd = 'echo $(hadoop classpath)'
        with cluster_context.get_instance(SPARK_HISTORY_SERVER).remote() as r:
            result = r.execute_command(cmd, run_as_root=True, timeout=600)
        return result[1].replace('\n', '').split(':')

    def _hive_libs_path(self, cluster_context):
        cmd = "find %s -name '*.jar'" % (
            self._hive(cluster_context).home_dir(cluster_context) + '/lib')
        with cluster_context.get_instance(hive.HIVE_METASTORE).remote() as r:
            result = r.execute_command(cmd, run_as_root=True, timeout=600)
        return [x for x in list(result[1].split('\n')) if x]

    def _assembly_jar_path(self, cluster_context):
        cmd = "find %s -name 'spark-assembly*.jar'" % (
            self.home_dir(cluster_context) + '/lib')
        with cluster_context.get_instance(SPARK_HISTORY_SERVER).remote() as r:
            result = r.execute_command(cmd, run_as_root=True, timeout=600)
        if result[1]:
            return result[1].strip()
        else:
            raise Exception("no spark-assembly lib found!")

    def _hive_datanucleus_libs_path(self, cluster_context):
        cmd = "find %s -name 'datanucleus-*.jar'" % (
            self._hive(cluster_context).home_dir(cluster_context) + '/lib')
        with cluster_context.get_instance(hive.HIVE_METASTORE).remote() as r:
            result = r.execute_command(cmd, run_as_root=True, timeout=600)
        return [x for x in list(result[1].split('\n')) if x]

    # hive installed service instance
    def _hive(self, cluster_context):
        hive_instance = cluster_context.get_instance(hive.HIVE_SERVER_2)
        if not hive_instance:
            return None
        hive_version = cluster_context.get_chosen_service_version('Hive')
        return cluster_context._find_service_instance('Hive', hive_version)

    # hbase installed service instance
    def _hbase(self, cluster_context):
        hbase_instance = cluster_context.get_instance(hbase.HBASE_MASTER)
        if not hbase_instance:
            return None
        hbase_version = cluster_context.get_chosen_service_version('HBase')
        return cluster_context._find_service_instance('HBase', hbase_version)

    def _get_hbase_version(self, cluster_context):
        return (self._hbase(cluster_context).version
                if self._hbase(cluster_context) else None)

    def _get_hive_version(self, cluster_context):
        return (self._hive(cluster_context).version
                if self._hive(cluster_context) else None)

    # hue installed service instance
    def _hue(self, cluster_context):
        hue_instance = cluster_context.get_instance('Hue')
        if not hue_instance:
            return None
        hue_version = cluster_context.get_chosen_service_version('Hue')
        return cluster_context._find_service_instance('Hue', hue_version)

    def _copy_jar_from_hue(self, cluster_context):
        if not self._hue(cluster_context):
            return
        jar_path = "%s/apps/spark/java-lib/javax.servlet-api-*.jar" % \
                   self._hue(cluster_context).home_dir(cluster_context)
        path = '%s/lib/' % self.home_dir(cluster_context) + self.SERVLET_JAR
        with cluster_context.get_instance('Hue').remote() as r1:
            for instance in cluster_context.get_instances(SPARK_SLAVE):
                with instance.remote() as r2:
                    mfs.exchange(r1, r2, jar_path, path, 'mapr')


class SparkOnYarnV161(SparkOnYarn):
    def __init__(self):
        super(SparkOnYarnV161, self).__init__()
        self._version = '1.6.1'
        self._dependencies = [('mapr-spark', self.version)]


class SparkOnYarnV201(SparkOnYarn):
    def __init__(self):
        super(SparkOnYarnV201, self).__init__()
        self._version = '2.0.1'
        self._dependencies = [('mapr-spark', self.version)]

    def get_config_files(self, cluster_context, configs, instance=None):
        hbase_version = self._get_hbase_version(cluster_context)
        hive_version = self._get_hive_version(cluster_context)
        # spark-env-sh
        template = 'plugins/mapr/services/' \
                   'spark/resources/spark-env.template'
        env_sh = bcf.TemplateFile('spark-env.sh')
        env_sh.remote_path = self.conf_dir(cluster_context)
        env_sh.parse(utils.get_file_text(template, 'sahara_plugin_mapr'))
        env_sh.add_property('version', self.version)

        # spark-defaults
        conf = bcf.PropertiesFile('spark-defaults.conf', separator=' ')
        conf.remote_path = self.conf_dir(cluster_context)
        if instance:
            conf.fetch(instance)

        # compatibility.version
        versions = bcf.PropertiesFile('compatibility.version')
        versions.remote_path = self.home_dir(cluster_context) + '/mapr-util'
        if instance:
            versions.fetch(instance)

        if hive_version:
            versions.add_property('hive_versions',
                                  hive_version + '.0')
            conf.add_properties(self._hive_properties(cluster_context))
        if hbase_version:
            versions.add_property('hbase_versions', hbase_version)
            conf.add_property('spark.executor.extraClassPath',
                              '%s/lib/*' % self._hbase(cluster_context)
                              .home_dir(cluster_context))
        return [conf, versions, env_sh]

    def post_start(self, cluster_context, instances):
        self._create_hadoop_spark_dirs(cluster_context)
        self._copy_hive_site(cluster_context)
        self._copy_hbase_site(cluster_context)



# Copyright (c) 2015, MapR Technologies
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sahara.plugins.provisioning as p
import sahara.plugins.utils as utils
from sahara_plugin_mapr.i18n import _
import sahara_plugin_mapr.plugins.mapr.domain.configuration_file as cf
import sahara_plugin_mapr.plugins.mapr.domain.node_process as np
import sahara_plugin_mapr.plugins.mapr.domain.service as s
import sahara_plugin_mapr.plugins.mapr.services.mysql.mysql as mysql
import sahara_plugin_mapr.plugins.mapr.util.event_log as el
import sahara_plugin_mapr.plugins.mapr.util.maprfs_helper as mfs

SENTRY = np.NodeProcess(
    name='sentry',
    ui_name='Sentry',
    package='mapr-sentry'
)

SENTRY_MODE_CONFIG_NAME = 'Sentry storage mode'
FILE_STORAGE_SENTRY_MODE = 'File-base storage'
DB_STORAGE_SENTRY_MODE = 'DB-based storage'


class Sentry(s.Service):
    SENTRY_STORAGE_MODE = p.Config(
        name=SENTRY_MODE_CONFIG_NAME,
        applicable_target='Sentry',
        scope='cluster',
        config_type="dropdown",
        config_values=[(v, v) for v in
                       (FILE_STORAGE_SENTRY_MODE, DB_STORAGE_SENTRY_MODE)],
        priority=1,
        description=_(
            'Specifies Sentry storage mode.')
    )
    GLOBAL_POLICY_FILE = '/user/mapr/sentry/global-policy.ini'

    def __init__(self):
        super(Sentry, self).__init__()
        self._name = 'sentry'
        self._ui_name = 'Sentry'
        self._node_processes = [SENTRY]
        self._priority = 2

    def get_configs(self):
        return [Sentry.SENTRY_STORAGE_MODE]

    def get_config_files(self, cluster_context, configs, instance=None):
        sentry_default = \
            'plugins/mapr/services/sentry/resources/sentry-default.xml'
        global_policy_template = \
            'plugins/mapr/services/sentry/resources/global-policy.ini'
        sentry_site = cf.HadoopXML('sentry-site.xml')
        sentry_site.remote_path = self.conf_dir(cluster_context)
        if instance:
            sentry_site.fetch(instance)
        sentry_site.load_properties(configs)
        sentry_mode = configs[self.SENTRY_STORAGE_MODE.name]
        sentry_site.parse(utils.get_file_text(sentry_default,
                                              'sahara_plugin_mapr'))
        sentry_site.add_properties(
            self._get_sentry_site_props(cluster_context, sentry_mode))
        global_policy = cf.TemplateFile('global-policy.ini')
        global_policy.remote_path = self.conf_dir(cluster_context)
        global_policy.parse(utils.get_file_text(global_policy_template,
                                                'sahara_plugin_mapr'))
        return [sentry_site, global_policy]

    def _get_jdbc_uri(self, cluster_context):
        jdbc_uri = ('jdbc:mysql://%(db_host)s:%(db_port)s/%(db_name)s?'
                    'createDatabaseIfNotExist=true')
        jdbc_args = {
            'db_host': mysql.MySQL.get_db_instance(
                cluster_context).internal_ip,
            'db_port': mysql.MySQL.MYSQL_SERVER_PORT,
            'db_name': mysql.MySQL.SENTRY_SPECS.db_name,
        }
        return jdbc_uri % jdbc_args

    def _get_sentry_site_props(self, cluster_context, setry_mode):
        sentry_specs = mysql.MySQL.SENTRY_SPECS
        if setry_mode == FILE_STORAGE_SENTRY_MODE:
            return {
                'sentry.hive.provider.backend':
                    'org.apache.sentry.provider'
                    '.file.SimpleFileProviderBackend',
                'sentry.hive.provider.resource':
                    'maprfs:///' + self.GLOBAL_POLICY_FILE,
            }
        if setry_mode == DB_STORAGE_SENTRY_MODE:
            return {
                'sentry.hive.provider.backend':
                    'org.apache.sentry.provider.db.SimpleDBProviderBackend',
                'sentry.store.jdbc.url': self._get_jdbc_uri(cluster_context),
                'sentry.store.jdbc.driver': mysql.MySQL.DRIVER_CLASS,
                'sentry.store.jdbc.user': sentry_specs.user,
                'sentry.store.jdbc.password': sentry_specs.password,
            }

    def _init_db_schema(self, cluster_context):
        instance = cluster_context.get_instance(SENTRY)
        cmd = '%(home)s/bin/sentry --command schema-tool' \
              ' --conffile %(home)s/conf/sentry-site.xml' \
              ' --dbType mysql --initSchema' % {
                  'home': self.home_dir(cluster_context)}

        @el.provision_event(name=_("Init Sentry DB schema"), instance=instance)
        def decorated():
            with instance.remote() as r:
                r.execute_command(cmd, run_as_root=True)

        decorated()

    def post_start(self, cluster_context, instances):
        sentry_host = cluster_context.get_instance(SENTRY)
        source = self.conf_dir(cluster_context) + '/global-policy.ini'
        with sentry_host.remote() as r:
            mfs.mkdir(r, '/user/mapr/sentry', run_as='mapr')
            mfs.chmod(r, '/user/mapr/sentry', 777, run_as='mapr')
            mfs.copy_from_local(r, source, self.GLOBAL_POLICY_FILE,
                                hdfs_user='mapr')

    def _copy_warden_conf(self, cluster_context):
        sentry_host = cluster_context.get_instance(SENTRY)
        cmd = 'sudo -u mapr cp %s/conf.d/warden.sentry.conf' \
              ' /opt/mapr/conf/conf.d/' % self.home_dir(cluster_context)
        with sentry_host.remote() as r:
            r.execute_command(cmd)

    def post_install(self, cluster_context, instances):
        self._set_service_dir_owner(cluster_context, instances)
        if cluster_context._get_cluster_config_value(
                self.SENTRY_STORAGE_MODE) == DB_STORAGE_SENTRY_MODE:
            self._init_db_schema(cluster_context)
            self._copy_warden_conf(cluster_context)

    def supports(self, service, mode):
        "return True is Sentry supports integration"
        service = service.name + '-' + service.version
        return self.SENTRY_SUPPORT_MATRIX[mode][service]


class SentryV16(Sentry):
    SENTRY_SUPPORT_MATRIX = {
        DB_STORAGE_SENTRY_MODE: {
            'hive-1.2': True,
            'hive-2.0': True,
            'impala-2.2.0': True,
            'impala-2.5.0': True
        },
        FILE_STORAGE_SENTRY_MODE: {
            'hive-1.2': True,
            'hive-2.0': True,
            'impala-2.2.0': True,
            'impala-2.5.0': True
        },
    }

    def __init__(self):
        super(SentryV16, self).__init__()
        self._version = '1.6.0'



# Copyright (c) 2015, MapR Technologies
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import sahara_plugin_mapr.plugins.mapr.domain.node_process as np
import sahara_plugin_mapr.plugins.mapr.domain.service as s
import sahara_plugin_mapr.plugins.mapr.util.validation_utils as vu


HTTP_FS = np.NodeProcess(
    name='httpfs',
    ui_name='HTTPFS',
    package='mapr-httpfs',
    open_ports=[14000]
)


class HttpFS(s.Service):
    def __init__(self):
        super(HttpFS, self).__init__()
        self._name = 'httpfs'
        self._ui_name = 'HttpFS'
        self._version = '1.0'
        self._node_processes = [HTTP_FS]
        self._cluster_defaults = ['httpfs-default.json']
        self._validation_rules = [vu.exactly(1, HTTP_FS)]

    def post_install(self, cluster_context, instances):
        self._set_service_dir_owner(cluster_context, instances)



# Copyright (c) 2016 Intel Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from oslo_utils import uuidutils

from sahara.plugins import castellan_utils as key_manager
from sahara.plugins import conductor
from sahara.plugins import context

MAPR_USER_PASSWORD = 'mapr_password'


def delete_password(cluster, pw_name):
    """delete the named password from the key manager

    This function will lookup the named password in the cluster entry
    and delete it from the key manager.

    :param cluster: The cluster record containing the password
    :param pw_name: The name associated with the password
    """
    ctx = context.ctx()
    cluster = conductor.cluster_get(ctx, cluster.id)
    key_id = cluster.extra.get(pw_name) if cluster.extra else None
    if key_id is not None:
        key_manager.delete_key(key_id, ctx)


def get_password(cluster, pw_name):
    """return a password for the named entry

    This function will return, or create and return, a password for the
    named entry. It will store the password in the key manager and use
    the ID in the database entry.

    :param cluster: The cluster record containing the password
    :param pw_name: The entry name associated with the password
    :returns: The cleartext password
    """
    ctx = context.ctx()
    cluster = conductor.cluster_get(ctx, cluster.id)
    passwd = cluster.extra.get(pw_name) if cluster.extra else None
    if passwd:
        return key_manager.get_secret(passwd, ctx)

    passwd = uuidutils.generate_uuid()
    extra = cluster.extra.to_dict() if cluster.extra else {}
    extra[pw_name] = key_manager.store_secret(passwd, ctx)
    cluster = conductor.cluster_update(ctx, cluster, {'extra': extra})
    return passwd


def get_mapr_password(cluster):
    return get_password(cluster, MAPR_USER_PASSWORD)



# Copyright (c) 2015, MapR Technologies
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import abc


class AbstractClusterContext(object, metaclass=abc.ABCMeta):
    @abc.abstractproperty
    def mapr_home(self):
        return

    @abc.abstractproperty
    def configure_sh_path(self):
        return

    @abc.abstractproperty
    def configure_sh(self):
        return

    @abc.abstractproperty
    def hadoop_version(self):
        return

    @abc.abstractproperty
    def hadoop_home(self):
        return

    @abc.abstractproperty
    def hadoop_lib(self):
        return

    @abc.abstractproperty
    def hadoop_conf(self):
        return

    @abc.abstractproperty
    def cluster(self):
        return

    @abc.abstractproperty
    def name_node_uri(self):
        return

    @abc.abstractproperty
    def resource_manager_uri(self):
        return

    @abc.abstractproperty
    def oozie_server_uri(self):
        return

    @abc.abstractproperty
    def oozie_server(self):
        return

    @abc.abstractproperty
    def oozie_http(self):
        return

    @abc.abstractproperty
    def cluster_mode(self):
        return

    @abc.abstractproperty
    def is_node_aware(self):
        return

    @abc.abstractproperty
    def some_instance(self):
        return

    @abc.abstractproperty
    def distro(self):
        return

    @abc.abstractproperty
    def mapr_db(self):
        return

    @abc.abstractmethod
    def filter_instances(self, instances, node_process=None, service=None):
        return

    @abc.abstractmethod
    def removed_instances(self, node_process=None, service=None):
        return

    @abc.abstractmethod
    def added_instances(self, node_process=None, service=None):
        return

    @abc.abstractmethod
    def changed_instances(self, node_process=None, service=None):
        return

    @abc.abstractmethod
    def existing_instances(self, node_process=None, service=None):
        return

    @abc.abstractproperty
    def should_be_restarted(self):
        return

    @abc.abstractproperty
    def mapr_repos(self):
        return

    @abc.abstractproperty
    def is_prebuilt(self):
        return

    @abc.abstractproperty
    def local_repo(self):
        return

    @abc.abstractproperty
    def required_services(self):
        return

    @abc.abstractproperty
    def all_services(self):
        return

    @abc.abstractproperty
    def mapr_version(self):
        return

    @abc.abstractproperty
    def ubuntu_base_repo(self):
        return

    @abc.abstractproperty
    def ubuntu_ecosystem_repo(self):
        return

    @abc.abstractproperty
    def centos_base_repo(self):
        return

    @abc.abstractproperty
    def centos_ecosystem_repo(self):
        return



# Copyright (c) 2015, MapR Technologies
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import collections

from oslo_config import cfg

import sahara.plugins.exceptions as e
from sahara.plugins import topology_helper as th
import sahara.plugins.utils as u
from sahara_plugin_mapr.i18n import _
import sahara_plugin_mapr.plugins.mapr.abstract.cluster_context as cc
import sahara_plugin_mapr.plugins.mapr.domain.configuration_file as bcf
import sahara_plugin_mapr.plugins.mapr.domain.distro as distro
import sahara_plugin_mapr.plugins.mapr.services.management.management as mng
import sahara_plugin_mapr.plugins.mapr.services.maprfs.maprfs as mfs
import sahara_plugin_mapr.plugins.mapr.services.oozie.oozie as oozie
from sahara_plugin_mapr.plugins.mapr.services.swift import swift
import sahara_plugin_mapr.plugins.mapr.services.yarn.yarn as yarn
import sahara_plugin_mapr.plugins.mapr.util.general as g
import sahara_plugin_mapr.plugins.mapr.util.service_utils as su

CONF = cfg.CONF
CONF.import_opt("enable_data_locality", "sahara.topology.topology_helper")


class BaseClusterContext(cc.AbstractClusterContext):
    ubuntu_base = 'http://package.mapr.com/releases/v%s/ubuntu/ mapr optional'
    centos_base = 'http://package.mapr.com/releases/v%s/redhat/'

    def __init__(self, cluster, version_handler, added=None, removed=None):
        self._cluster = cluster
        self._distro = None
        self._distro_version = None
        self._all_services = version_handler.get_services()
        self._required_services = version_handler.get_required_services()
        self._cluster_services = None
        self._mapr_home = '/opt/mapr'
        self._name_node_uri = 'maprfs:///'
        self._cluster_mode = None
        self._node_aware = None
        self._oozie_server_uri = None
        self._oozie_server = None
        self._oozie_http = None
        self._some_instance = None
        self._configure_sh_path = None
        self._configure_sh = None
        self._mapr_db = None
        self._hadoop_home = None
        self._hadoop_version = None
        self._added_instances = added or []
        self._removed_instances = removed or []
        self._changed_instances = (
            self._added_instances + self._removed_instances)
        self._existing_instances = [i for i in self.get_instances()
                                    if i not in self._changed_instances]
        self._restart = collections.defaultdict(list)
        self._ubuntu_base_repo = None
        self._ubuntu_ecosystem_repo = None
        self._centos_base_repo = None
        self._centos_ecosystem_repo = None
        self._repos = {}
        self._is_prebuilt = None
        self._local_repo = '/opt/mapr-repository'
        self._mapr_version = None

    @property
    def cluster(self):
        return self._cluster

    @property
    def cluster_services(self):
        if not self._cluster_services:
            self._cluster_services = self.get_cluster_services()
        return self._cluster_services

    @property
    def required_services(self):
        return self._required_services

    @property
    def all_services(self):
        return self._all_services

    @property
    def mapr_home(self):
        return self._mapr_home

    @property
    def hadoop_version(self):
        return self._hadoop_version

    @property
    def hadoop_home(self):
        if not self._hadoop_home:
            f = '%(mapr_home)s/hadoop/hadoop-%(hadoop_version)s'
            args = {
                'mapr_home': self.mapr_home,
                'hadoop_version': self.hadoop_version,
            }
            self._hadoop_home = f % args
        return self._hadoop_home

    @property
    def name_node_uri(self):
        return self._name_node_uri

    @property
    def oozie_server_uri(self):
        if not self._oozie_server_uri:
            oozie_http = self.oozie_http
            url = 'http://%s/oozie' % oozie_http if oozie_http else None
            self._oozie_server_uri = url
        return self._oozie_server_uri

    @property
    def oozie_server(self):
        if not self._oozie_server:
            self._oozie_server = self.get_instance(oozie.OOZIE)
        return self._oozie_server

    @property
    def oozie_http(self):
        if not self._oozie_http:
            oozie_server = self.oozie_server
            ip = oozie_server.management_ip if oozie_server else None
            self._oozie_http = '%s:11000' % ip if ip else None
        return self._oozie_http

    @property
    def cluster_mode(self):
        return self._cluster_mode

    @property
    def is_node_aware(self):
        return self._node_aware and CONF.enable_data_locality

    @property
    def some_instance(self):
        if not self._some_instance:
            self._some_instance = self.cluster.node_groups[0].instances[0]
        return self._some_instance

    @property
    def distro_version(self):
        if not self._distro_version:
            self._distro_version = distro.get_version(self.some_instance)
        return self._distro_version

    @property
    def distro(self):
        if not self._distro:
            self._distro = distro.get(self.some_instance)
        return self._distro

    @property
    def mapr_db(self):
        if self._mapr_db is None:
            mapr_db = mfs.MapRFS.ENABLE_MAPR_DB_CONFIG
            mapr_db = self._get_cluster_config_value(mapr_db)
            self._mapr_db = '-noDB' if not mapr_db else ''
        return self._mapr_db

    @property
    def configure_sh_path(self):
        if not self._configure_sh_path:
            self._configure_sh_path = '%s/server/configure.sh' % self.mapr_home
        return self._configure_sh_path

    @property
    def configure_sh(self):
        if not self._configure_sh:
            f = ('%(script_path)s'
                 ' -N %(cluster_name)s'
                 ' -C %(cldbs)s'
                 ' -Z %(zookeepers)s'
                 ' -no-autostart -f %(m7)s')
            args = {
                'script_path': self.configure_sh_path,
                'cluster_name': self.cluster.name,
                'cldbs': self.get_cldb_nodes_ip(),
                'zookeepers': self.get_zookeeper_nodes_ip(),
                'm7': self.mapr_db
            }
            self._configure_sh = f % args
        return self._configure_sh

    def _get_cluster_config_value(self, config):
        cluster_configs = self.cluster.cluster_configs
        service = config.applicable_target
        name = config.name
        if service in cluster_configs and name in cluster_configs[service]:
            return cluster_configs[service][name]
        else:
            return config.default_value

    def get_node_processes(self):
        node_processes = []
        for ng in self.cluster.node_groups:
            for np in ng.node_processes:
                if np not in node_processes:
                    node_processes.append(self.get_node_process_by_name(np))
        return node_processes

    def get_node_process_by_name(self, name):
        for service in self.cluster_services:
            for node_process in service.node_processes:
                if node_process.ui_name == name:
                    return node_process

    def get_instances(self, node_process=None):
        if node_process is not None:
            node_process = su.get_node_process_name(node_process)
        return u.get_instances(self.cluster, node_process)

    def get_instance(self, node_process):
        node_process_name = su.get_node_process_name(node_process)
        instances = u.get_instances(self.cluster, node_process_name)
        return instances[0] if instances else None

    def get_instances_ip(self, node_process):
        return [i.internal_ip for i in self.get_instances(node_process)]

    def get_instance_ip(self, node_process):
        i = self.get_instance(node_process)
        return i.internal_ip if i else None

    def get_zookeeper_nodes_ip_with_port(self, separator=','):
        return separator.join(['%s:%s' % (ip, mng.ZK_CLIENT_PORT)
                               for ip in self.get_instances_ip(mng.ZOOKEEPER)])

    def check_for_process(self, instance, process):
        return su.has_node_process(instance, process)

    def get_services_configs_dict(self, services=None):
        if not services:
            services = self.cluster_services
        result = dict()
        for service in services:
            result.update(service.get_configs_dict())
        return result

    def get_chosen_service_version(self, service_name):
        service_configs = self.cluster.cluster_configs.get(service_name, None)
        if not service_configs:
            return None
        return service_configs.get('%s Version' % service_name, None)

    def get_cluster_services(self, node_group=None):
        node_processes = None

        if node_group:
            node_processes = node_group.node_processes
        else:
            node_processes = [np for ng in self.cluster.node_groups
                              for np in ng.node_processes]
            node_processes = g.unique_list(node_processes)
        services = g.unique_list(node_processes, self.get_service)

        return services + [swift.Swift()]

    def get_service(self, node_process):
        ui_name = self.get_service_name_by_node_process(node_process)
        if ui_name is None:
            raise e.PluginInvalidDataException(
                _('Service not found in services list'))
        version = self.get_chosen_service_version(ui_name)
        service = self._find_service_instance(ui_name, version)
        if service is None:
            raise e.PluginInvalidDataException(_('Can not map service'))
        return service

    def _find_service_instance(self, ui_name, version):
        # if version is None, the latest service version is returned
        for service in self.all_services[::-1]:
            if service.ui_name == ui_name:
                if version is not None and service.version != version:
                    continue
                return service

    def get_service_name_by_node_process(self, node_process):
        node_process_name = su.get_node_process_name(node_process)
        for service in self.all_services:
            service_node_processes = [np.ui_name
                                      for np in service.node_processes]
            if node_process_name in service_node_processes:
                return service.ui_name

    def get_instances_count(self, node_process=None):
        if node_process is not None:
            node_process = su.get_node_process_name(node_process)
        return u.get_instances_count(self.cluster, node_process)

    def get_node_groups(self, node_process=None):
        if node_process is not None:
            node_process = su.get_node_process_name(node_process)
        return u.get_node_groups(self.cluster, node_process)

    def get_cldb_nodes_ip(self, separator=','):
        return separator.join(self.get_instances_ip(mfs.CLDB))

    def get_zookeeper_nodes_ip(self, separator=','):
        return separator.join(
            self.get_instances_ip(mng.ZOOKEEPER))

    def get_resourcemanager_ip(self):
        return self.get_instance_ip(yarn.RESOURCE_MANAGER)

    def get_historyserver_ip(self):
        return self.get_instance_ip(yarn.HISTORY_SERVER)

    def has_control_nodes(self, instances):
        for inst in instances:
            zookeepers = self.check_for_process(inst, mng.ZOOKEEPER)
            cldbs = self.check_for_process(inst, mfs.CLDB)
            if zookeepers or cldbs:
                return True
        return False

    def is_present(self, service):
        is_service_subclass = lambda s: isinstance(s, service.__class__)
        return any(is_service_subclass(s) for s in self.cluster_services)

    def filter_instances(self, instances, node_process=None, service=None):
        if node_process:
            return su.filter_by_node_process(instances, node_process)
        if service:
            return su.filter_by_service(instances, service)
        return list(instances)

    def removed_instances(self, node_process=None, service=None):
        instances = self._removed_instances
        return self.filter_instances(instances, node_process, service)

    def added_instances(self, node_process=None, service=None):
        instances = self._added_instances
        return self.filter_instances(instances, node_process, service)

    def changed_instances(self, node_process=None, service=None):
        instances = self._changed_instances
        return self.filter_instances(instances, node_process, service)

    def existing_instances(self, node_process=None, service=None):
        instances = self._existing_instances
        return self.filter_instances(instances, node_process, service)

    @property
    def should_be_restarted(self):
        return self._restart

    @property
    def mapr_repos(self):
        if not self._repos:
            self._repos = {
                "ubuntu_mapr_base_repo": self.ubuntu_base_repo,
                "ubuntu_mapr_ecosystem_repo": self.ubuntu_ecosystem_repo,
                "centos_mapr_base_repo": self.centos_base_repo,
                "centos_mapr_ecosystem_repo": self.centos_ecosystem_repo,
            }
        return self._repos

    @property
    def local_repo(self):
        return self._local_repo

    @property
    def is_prebuilt(self):
        if self._is_prebuilt is None:
            self._is_prebuilt = g.is_directory(
                self.some_instance, self.local_repo)
        return self._is_prebuilt

    @property
    def mapr_version(self):
        return self._mapr_version

    @property
    def ubuntu_base_repo(self):
        default_value = self._ubuntu_base_repo or self.ubuntu_base % self\
            .mapr_version
        return self.cluster.cluster_configs.get(
            'general', {}).get('Ubuntu base repo', default_value)

    @property
    def ubuntu_ecosystem_repo(self):
        default_value = self._ubuntu_ecosystem_repo
        return self.cluster.cluster_configs.get(
            'general', {}).get('Ubuntu ecosystem repo', default_value)

    @property
    def centos_base_repo(self):
        default_value = self._centos_base_repo or self.centos_base % self\
            .mapr_version
        return self.cluster.cluster_configs.get(
            'general', {}).get('CentOS base repo', default_value)

    @property
    def centos_ecosystem_repo(self):
        default_value = self._centos_ecosystem_repo
        return self.cluster.cluster_configs.get(
            'general', {}).get('CentOS ecosystem repo', default_value)

    def get_configuration(self, node_group):
        services = self.get_cluster_services(node_group)
        user_configs = node_group.configuration()
        default_configs = self.get_services_configs_dict(services)
        return u.merge_configs(default_configs, user_configs)

    def get_config_files(self, node_group):
        services = self.get_cluster_services(node_group)
        configuration = self.get_configuration(node_group)
        instance = node_group.instances[0]

        config_files = []
        for service in services:
            service_conf_files = service.get_config_files(
                cluster_context=self,
                configs=configuration[service.ui_name],
                instance=instance,
            )
            for conf_file in service_conf_files:
                file_atr = bcf.FileAttr(conf_file.remote_path,
                                        conf_file.render(), conf_file.mode,
                                        conf_file.owner)
                config_files.append(file_atr)

        return config_files

    @property
    def topology_map(self):
        return th.generate_topology_map(self.cluster, self.is_node_aware)



# Copyright (c) 2015, MapR Technologies
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from sahara_plugin_mapr.plugins.mapr.domain import distro
import sahara_plugin_mapr.tests.unit.base as b


class TestDistro(b.SaharaTestCase):
    def __init__(self, *args, **kwds):
        super(TestDistro, self).__init__(*args, **kwds)
        self.install_cmd = 'foo_bar'
        self.separator = '-'
        self.distro = distro.Distro('foo', 'foo', self.install_cmd,
                                    self.separator)

    def test_create_install_cmd(self):
        pkgs = [('foo',), ('bar', 'version')]
        cmd = self.distro.create_install_cmd(pkgs)
        self.assertIsNotNone(cmd)
        parts = cmd.split(' ')
        self.assertEqual(self.install_cmd, parts[0])
        self.assertEqual('foo', parts[1])
        self.assertEqual('bar', parts[2].split(self.separator)[0])
        self.assertEqual('version*', parts[2].split(self.separator)[1])
